# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import numba as nb
import numpy as np
import scipy as sp


def norm_inf(x):
    """Computes the infinity norm of a vector, i.e., maximum absolute value.

    Parameters
    ----------
    x : ndarray
        Vector to compute infinity norm of.

    Returns
    -------
    float
        Infinity norm of x.
    """
    return abs(max(x.max(initial=0.0), -x.min(initial=0.0)))


def inp(x, y):
    """Computes the standard inner product between two vectors.

    Parameters
    ----------
    x : ndarray
        First vector to compute inner product with.
    y : ndarray
        Second vector to compute inner product with. Should be the same type and
        dimension as x.

    Returns
    -------
    float
        Standard inner product between x and y.
    """
    x_view = x.view(dtype=np.float64).reshape(1, -1)
    y_view = y.view(dtype=np.float64).reshape(-1, 1).conj()
    return (x_view @ y_view)[0, 0]


def cho_fact(A, increment_diag=True):
    """Perform a Cholesky decomposition on a positive definite matrix. Increment
    diagonals by a small amount if Cholesky decomposition fails until the factorization
    succeeds.

    Parameters
    ----------
    A : ndarray
        Symmetric matrix to compute Cholesky decomposition on. It should be assumed that
        A is positive definite (up to some numerical errors).

    Returns
    -------
    tuple(ndarray, bool)
        Cholesky decomposition of A from SciPy.
    """
    diag_incr = np.finfo(A.dtype).eps
    while True:
        try:
            fact = sp.linalg.cho_factor(A, check_finite=False)
            return fact
        except np.linalg.LinAlgError:
            if not increment_diag:
                return None
            A.flat[:: A.shape[0] + 1] += diag_incr
            diag_incr *= 1e1


def cho_solve(fact, b):
    """Factor solve for Cholesky factorization of A with vectors b

    Parameters
    ----------
    fact : tuple[ndarray, bool]
        Cholesky decomposition of a (n, n) matrix A returned from SciPy.
    b : ndarray
        An (n, 1) vector, or (n, k) list of vectors, to solve linear systems with.

    Returns
    -------
    ndarray
        A (n, k) vector containing the solutions to the linear systems.
    """
    return sp.linalg.cho_solve(fact, b, check_finite=False)


def congr_multi(out, A, X, work, B=None):
    """Perform a congruence transform on a list of matrices.

        Xi --> A Xi A'    for    i = 1,...,k    if B = None
        Xi --> A Xi B'    for    i = 1,...,k    otherwise

    Parameters
    ----------
    out : ndarray
        Preallocated (k, m, m) array to store the output in.
    A : ndarray
        The (m, n) matrix to left-multiply with.
    X : ndarray
        The (k, n, n) list of matrices to perform a congruence transform on.
    work : ndarray
        Preallocated (k, m, n) array to do work with.
    B : ndarray, optional
        The (m, n) matrix to right-multiply with. Default is A.
    """
    if B is None:
        B = A

    # Performs congruence A X_i B' for i = i,...,n
    n, m = A.shape
    np.matmul(A, X, out=work)
    np.matmul(work.reshape((-1, m)), B.conj().T, out=out.reshape((-1, n)))
    return out


def scale_axis(A, scale_rows=None, scale_cols=None):
    if sp.sparse.issparse(A):
        A_coo = A.tocoo()
        if scale_rows is not None:
            A_coo.data *= np.take(scale_rows.ravel(), A_coo.row)
        if scale_cols is not None:
            A_coo.data *= np.take(scale_cols.ravel(), A_coo.col)
        return A_coo
    else:
        if scale_rows is not None:
            A *= scale_rows.reshape((-1, 1))
        if scale_cols is not None:
            A *= scale_cols.reshape((1, -1))
        return A


def abs_max(A, axis):
    if sp.sparse.issparse(A):
        A = A.copy()
        A.data = np.abs(A.data)
        return A.max(axis=axis).toarray().reshape(-1)
    else:
        return np.maximum(A.max(axis=axis, initial=0.0), -A.min(axis=axis, initial=0.0))


def is_full_col_rank(A, tol=1e-8):
    if A.size == 0:
        return True
    AA = A.T @ A
    if sp.sparse.issparse(AA):
        AA = AA.toarray()
    eigs = np.linalg.eigvalsh(AA)
    return all(eigs > tol)


def dense_dot_x(A, B):
    if sp.sparse.issparse(B):
        return dense_dot_sparse(A, B.col, B.row, B.data, B.shape)
    else:
        return A @ B


def x_dot_dense(A, B):
    return dense_dot_x(B.T, A.T).T


@nb.njit(cache=True, parallel=True)
def dense_dot_sparse(A, B_col, B_row, B_val, B_shape):
    A_dot_B = np.zeros((A.shape[0], B_shape[1]))
    for j in nb.prange(A.shape[0]):
        A_dot_B[j, :] = 0.0
        for i in range(B_val.size):
            A_dot_B[j, B_col[i]] += A[j, B_row[i]] * B_val[i]
    return A_dot_B



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import numpy as np
import scipy as sp


class Vector:
    """Base class for a vector"""

    def __init__(self):
        self.vec = None

    def __iadd__(self, other):
        if self.vec.size > 0:
            self.vec[:] = sp.linalg.blas.daxpy(other.vec, self.vec, a=1)
        return self

    def __isub__(self, other):
        if self.vec.size > 0:
            self.vec[:] = sp.linalg.blas.daxpy(other.vec, self.vec, a=-1)
        return self

    def __imul__(self, a):
        self.vec *= a
        return self

    def axpy(self, a, other):
        if self.vec.size > 0:
            self.vec[:] = sp.linalg.blas.daxpy(other.vec, self.vec, a=a)
        return self

    def copy_from(self, other):
        if isinstance(other, np.ndarray):
            np.copyto(self.vec, other)
        else:
            np.copyto(self.vec, other.vec)
        return self

    def norm(self):
        return np.sqrt((self.vec.T @ self.vec)[0, 0])

    def inp(self, other):
        return (self.vec.T @ other.vec)[0, 0]

    def fill(self, a):
        self.vec.fill(a)
        return self


class Point(Vector):
    r"""A class for a vector containing the variables involved in a
    homogeneous self-dual embedding of a primal-dual conic program
    :math:`(x, y, z, s, \tau, \kappa)\in\mathbb{R}^n\times\mathbb{R}^p
    \times\mathbb{R}^q\times\mathbb{R}^q\times\mathbb{R}\times\mathbb{R}`.

    Parameters
    ----------
    model : :class:`~qics.Model`
        Model which specifies the conic program which this vector
        corresponds to.

    Attributes
    ----------
    vec : :class:`~numpy.ndarray`
        2D :obj:`~numpy.float64` array of size ``(n+p+2q+2, 1)``
        representing the full concatenated vector :math:`(x, y, z, s, \tau,
        \kappa)`.
    x : :class:`~numpy.ndarray`
        A :obj:`~numpy.ndarray.view` of ``vec`` of size ``(n, 1)``
        correpsonding to the primal variable :math:`x`.
    y : :class:`~numpy.ndarray`
        A :obj:`~numpy.ndarray.view` of ``vec`` of size ``(p, 1)``
        correpsonding to the dual variable :math:`y`.
    z : :class:`~qics.point.VecProduct`
        A :obj:`~numpy.ndarray.view` of ``vec`` of size ``(q, 1)``
        correpsonding to the dual variable :math:`z`.
    s : :class:`~qics.point.VecProduct`
        A :obj:`~numpy.ndarray.view` of ``vec`` of size ``(q, 1)``
        correpsonding to the primal variable :math:`s`.
    tau : :class:`~numpy.ndarray`
        A :obj:`~numpy.ndarray.view` of ``vec`` of size ``(1, 1)``
        correpsonding to the primal homogenizing variable :math:`\tau`.
    kap : :class:`~numpy.ndarray`
        A :obj:`~numpy.ndarray.view` of ``vec`` of size ``(1, 1)``
        correpsonding to the dual homogenizing variable :math:`\kappa`.
    """

    def __init__(self, model):
        (n, p, q) = (model.n, model.p, model.q)

        # Initialize vector
        self.vec = np.zeros((n + p + q + q + 2, 1))

        # Build views of vector
        self._xyz = PointXYZ(model, self.vec[: n + p + q])

        self.x = self._xyz.x
        self.y = self._xyz.y
        self.z = self._xyz.z
        self.s = VecProduct(model.cones, self.vec[n + p + q : n + p + q + q])
        self.tau = self.vec[n + p + q + q : n + p + q + q + 1]
        self.kap = self.vec[n + p + q + q + 1 : n + p + q + q + 2]

        return


class PointXYZ(Vector):
    """A class for a vector containing only the (x,y,z) variables invovled in a
    primal-dual conic program"""

    def __init__(self, model, vec=None):
        (n, p, q) = (model.n, model.p, model.q)

        # Initialize vector
        if vec is not None:
            # If view of vector is already given, use that
            assert vec.size == n + p + q
            self.vec = vec
        else:
            # Otherwise allocate a new vector
            self.vec = np.zeros((n + p + q, 1))

        # Build views of vector
        self.x = self.vec[:n]
        self.y = self.vec[n : n + p]
        self.z = VecProduct(model.cones, self.vec[n + p : n + p + q])

        return


class VecProduct(Vector):
    r"""A class for a Cartesian product of vectors corresponding to a list
    of cones :math:`\mathcal{K}_i`, i.e., :math:`s\in\mathbb{V}` where

    .. math::

        \mathbb{V} = \mathbb{V}_1 \times \mathbb{V}_2 \times \ldots \times
        \mathbb{V}_k,
    
    and :math:`\mathcal{K}_i \subset \mathbb{V}_i`. Each of these vector
    spaces :math:`\mathbb{V}_i` are themselves a Cartesian product of
    vector spaces

    .. math::

        \mathbb{V}_i = \mathbb{V}_{i,1} \times \mathbb{V}_{i,2} \times
         \ldots \times \mathbb{V}_{i,k_i},
    
    where :math:`\mathbb{V}_{i,j}` are defined as either the set of real 
    vectors :math:`\mathbb{R}^n`, symmetric matrices :math:`\mathbb{S}^n`,
    or Hermitian matrices :math:`\mathbb{H}^n`.

    Parameters
    ----------
    cones : :obj:`list` of :mod:`~qics.cones`
        List of cones defining a Cartesian product of vector spaces.
    vec : :class:`~numpy.ndarray`, optional
        If specified, then this class is initialized as a
        :obj:`~numpy.ndarray.view` of ``vec``. Otherwise, the class is
        initialized using a newly allocated :class:`~numpy.ndarray`.

    Attributes
    ----------
    vec : :class:`~numpy.ndarray`
        2D :obj:`~numpy.float64` array of size ``(q, 1)`` representing the
        full concatenated Cartesian product of vectors.
    mats : :obj:`list` of :obj:`list` of :class:`~numpy.ndarray`
        A nested list of :obj:`~numpy.ndarray.view` of ``vec`` where
        ``mats[i][j]`` returns the array corresponding to the vector space
        :math:`\mathbb{V}_{i,j}`. This attribute can also be called using
        ``__getitem__``, i.e., by directly calling ``self[i][j]``.
    vecs : :obj:`list` of :class:`~numpy.ndarray`
        A list of :obj:`~numpy.ndarray.view` of ``vec`` where ``vecs[i]``
        returns the array corresponding to the vector space
        :math:`\mathbb{V}_{i}` as a vectorized column vector.

    Examples
    --------
    Below we show an example of how to initialize a 
    :class:`~qics.point.VecProduct` and how to access the vectors
    corresponding to each cone and variable.

    >>> import qics
    >>> cones = [                                       \
    ...     qics.cones.PosSemidefinite(2),              \
    ...     qics.cones.QuantRelEntr(3, iscomplex=True)  \
    ... ]
    >>> x = qics.point.VecProduct(cones)
    >>> x[0][0]  # Matrix corresponding to PosSemidefinite cone
    array([[0., 0.],
           [0., 0.]])
    >>> x[1][0]  # Value corresponding to t of QuantRelEntr cone
    array([[0.]])
    >>> x[1][1]  # Matrix corresponding to X of QuantRelEntr cone
    array([[0.+0.j, 0.+0.j, 0.+0.j],
           [0.+0.j, 0.+0.j, 0.+0.j],
           [0.+0.j, 0.+0.j, 0.+0.j]])
    >>> x[1][2]  # Matrix corresponding to Y of QuantRelEntr cone
    array([[0.+0.j, 0.+0.j, 0.+0.j],
           [0.+0.j, 0.+0.j, 0.+0.j],
           [0.+0.j, 0.+0.j, 0.+0.j]])
    """

    def __init__(self, cones, vec=None):
        dims = []
        types = []
        dim = 0
        for cone_k in cones:
            dims.append(cone_k.dim)
            types.append(cone_k.type)
            dim += np.sum(cone_k.dim)

        # Initialize vector
        if vec is not None:
            # If view of vector is already given, use that
            assert vec.size == dim
            self.vec = vec
        else:
            # Otherwise allocate a new vector
            self.vec = np.zeros((dim, 1))

        # Build views of vector
        def vec_to_mat(vec, dim, type, t):
            if type == "r":
                # Real vector
                return vec[t : t + dim]
            elif type == "s":
                # Symmetric matrix
                n_k = int(np.sqrt(dim))
                return vec[t : t + dim].reshape((n_k, n_k))
            elif type == "h":
                # Hermitian matrix
                n_k = int(np.sqrt(dim // 2))
                return (
                    vec[t : t + dim]
                    .reshape((-1, 2))
                    .view(dtype=np.complex128)
                    .reshape(n_k, n_k)
                )

        self.vecs = []
        self.mats = []
        t = 0
        for dim_k, type_k in zip(dims, types):
            self.vecs.append(self.vec[t : t + np.sum(dim_k)])
            if isinstance(type_k, list):
                mats_k = []
                for dim_k_j, type_k_j in zip(dim_k, type_k):
                    mats_k.append(vec_to_mat(self.vec, dim_k_j, type_k_j, t))
                    t += dim_k_j
                self.mats.append(mats_k)
            else:
                self.mats.append(vec_to_mat(self.vec, dim_k, type_k, t))
                t += dim_k

    def __getitem__(self, key):
        return self.mats[key]



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import numpy as np
import scipy as sp

from qics._utils.gradient import D1_log, D2_log, scnd_frechet
from qics._utils.linalg import congr_multi, dense_dot_x, inp
from qics.cones.base import Cone, get_central_ray_entr
from qics.vectorize import vec_to_mat


class QuantEntr(Cone):
    r"""A class representing a (homogenized) quantum entropy cone

    .. math::

        \mathcal{QE}_{n} = \text{cl}\{ (t, u, X) \in \mathbb{R} \times
        \mathbb{R}_{++} \times \mathbb{H}^n_{++} : t \geq -u S(u^{-1}X) \},

    where

    .. math::

        S(X) = -\text{tr}[X \log(X)],

    is the quantum (von Neumann) entropy function.

    Parameters
    ----------
    n : :obj:`int`
        Dimension of the matrix :math:`X`.
    iscomplex : :obj:`bool`
        Whether the matrix :math:`X` is defined over :math:`\mathbb{H}^n`
        (``True``), or restricted to :math:`\mathbb{S}^n` (``False``). The
        default is ``False``.

    See also
    --------
    ClassEntr : (Homogenized) classical entropy cone
    QuantRelEntr : Quantum relative entropy cone

    Notes
    -----
    The epigraph of the quantum entropy can be obtained by enforcing the
    linear constraint :math:`u=1`.
    """

    def __init__(self, n, iscomplex=False):
        # Dimension properties
        self.n = n
        self.iscomplex = iscomplex

        self.nu = 2 + self.n  # Barrier parameter

        if iscomplex:
            self.dim = [1, 1, 2 * n * n]
            self.type = ["r", "r", "h"]
            self.dtype = np.complex128
        else:
            self.dim = [1, 1, n * n]
            self.type = ["r", "r", "s"]
            self.dtype = np.float64

        # Update flags
        self.feas_updated = False
        self.grad_updated = False
        self.hess_aux_updated = False
        self.invhess_aux_updated = False
        self.congr_aux_updated = False
        self.dder3_aux_updated = False

        return

    def get_iscomplex(self):
        return self.iscomplex

    def get_init_point(self, out):
        (t0, u0, x0) = get_central_ray_entr(self.n)

        point = [
            np.array([[t0]]),
            np.array([[u0]]),
            np.eye(self.n, dtype=self.dtype) * x0,
        ]

        self.set_point(point, point)

        out[0][:] = point[0]
        out[1][:] = point[1]
        out[2][:] = point[2]

        return out

    def get_feas(self):
        if self.feas_updated:
            return self.feas

        self.feas_updated = True

        (self.t, self.u, self.X) = self.primal

        # Check that u and X are positive (definite)
        self.Dx, self.Ux = np.linalg.eigh(self.X)

        if (self.u <= 0) or any(self.Dx <= 0):
            self.feas = False
            return self.feas

        # Check that t > -u S(X/u) = tr[X log(X)] - tr[X] log(u)
        self.trX = np.trace(self.X).real
        self.log_Dx = np.log(self.Dx)
        self.log_u = np.log(self.u[0, 0])

        entr_X = inp(self.Dx, self.log_Dx)
        entr_Xu = self.trX * self.log_u
        self.z = self.t[0, 0] - (entr_X - entr_Xu)

        self.feas = self.z > 0
        return self.feas

    def get_val(self):
        return -np.log(self.z) - np.sum(self.log_Dx) - self.log_u

    def update_grad(self):
        assert self.feas_updated
        assert not self.grad_updated

        # Compute gradients of quantum entropy
        # D_u S(u, X) = -tr[X] / u
        self.ui = np.reciprocal(self.u)
        self.DPhiu = -self.trX * self.ui
        # D_X S(u, X) = log(X) + (1 - log(u)) I
        log_X = (self.Ux * self.log_Dx) @ self.Ux.conj().T
        log_X = (log_X + log_X.conj().T) * 0.5
        self.DPhiX = log_X + np.eye(self.n) * (1.0 - self.log_u)

        # Compute X^-1
        inv_Dx = np.reciprocal(self.Dx)
        inv_X_rt2 = self.Ux * np.sqrt(inv_Dx)
        self.inv_X = inv_X_rt2 @ inv_X_rt2.conj().T

        # Compute gradient of barrier function
        self.zi = np.reciprocal(self.z)

        self.grad = [
            -self.zi,
            self.zi * self.DPhiu - self.ui,
            self.zi * self.DPhiX - self.inv_X,
        ]

        self.grad_updated = True

    def hess_prod_ip(self, out, H):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()

        (Ht, Hu, Hx) = H

        UxHxUx = self.Ux.conj().T @ Hx @ self.Ux

        # Hessian product of quantum entropy
        # D2_uu S(u, X) [Hu] = tr[X] Hu / u^2
        D2PhiuuH = self.trX * Hu * self.ui2
        # D2_uX S(u, X) [Hx] = -tr[Hx] / u
        D2PhiuXH = -np.trace(Hx).real * self.ui
        # D2_Xu S(u, X) [Hu] = -I Hu / u
        D2PhiXuH = -np.eye(self.n) * Hu * self.ui
        # D2_XX S(u, X) [Hx] =  Ux [log^[1](Dx) .* (Ux' Hx Ux)] Ux'
        D2PhiXXH = self.Ux @ (self.D1x_log * UxHxUx) @ self.Ux.conj().T

        # ======================================================================
        # Hessian products with respect to t
        # ======================================================================
        # D2_t F(t, u, X)[Ht, Hu, Hx]
        #         = (Ht - D_u S(u, X)[Hu] - D_X S(u, X)[Hx]) / z^2
        out[0][:] = (Ht - self.DPhiu * Hu - inp(self.DPhiX, Hx)) * self.zi2

        # ======================================================================
        # Hessian products with respect to u
        # ======================================================================
        # Hessian product of barrier function
        # D2_u F(t, u, X)[Ht, Hu, Hx]
        #         = -D2_t F(t, u, X)[Ht, Hu, Hx] * D_u S(u, X)
        #           + (D2_uu S(u, X)[Hu] + D2_uX S(u, X)[Hx]) / z
        #           + Hu / u^2
        out_u = -out[0] * self.DPhiu
        out_u += self.zi * (D2PhiuuH + D2PhiuXH)
        out_u += Hu * self.ui2
        out[1][:] = out_u

        # ======================================================================
        # Hessian products with respect to X
        # ======================================================================
        # D2_X F(t, u, X)[Ht, Hu, Hx]
        #         = -D2_t F(t, u, X)[Ht, Hu, Hx] * D_X S(u, X)
        #           + (D2_Xu S(u, X)[Hu] + D2_XX S(u, X)[Hx]) / z
        #           + X^-1 Hx X^-1
        out_X = -out[0] * self.DPhiX
        out_X += self.zi * (D2PhiXuH + D2PhiXXH)
        out_X += self.inv_X @ Hx @ self.inv_X
        out_X = (out_X + out_X.conj().T) * 0.5
        out[2][:] = out_X

        return out

    def hess_congr(self, A):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.congr_aux_updated:
            self.congr_aux(A)

        p = A.shape[0]
        lhs = np.zeros((p, sum(self.dim)))

        work1, work2, work3 = self.work1, self.work2, self.work3

        # ======================================================================
        # Hessian products with respect to t
        # ======================================================================
        # D2_t F(t, u, X)[Ht, Hu, Hx]
        #         = (Ht - D_u S(u, X)[Hu] - D_X S(u, X)[Hx]) / z^2
        DPhiX_vec = self.DPhiX.view(np.float64).reshape((-1, 1))
        out_t = self.At - (self.Ax_vec @ DPhiX_vec).ravel()
        out_t -= self.Au * self.DPhiu[0, 0]
        out_t *= self.zi2

        lhs[:, 0] = out_t

        # ======================================================================
        # Hessian products with respect to u
        # ======================================================================
        # Hessian products for quantum entropy
        # D2_uu S(u, X) [Hu] = tr[X] Hu / u^2
        D2PhiuuH = (self.trX * self.ui2) * self.Au
        # D2_uX S(u, X) [Hx] = -tr[Hx] / u
        D2PhiuXH = np.trace(self.Ax, axis1=1, axis2=2).real * self.ui

        # Hessian product of barrier function
        # D2_u F(t, u, X)[Ht, Hu, Hx]
        #         = -D2_t F(t, u, X)[Ht, Hu, Hx] * D_u S(u, X)
        #           + (D2_uu S(u, X)[Hu] + D2_uX S(u, X)[Hx]) / z
        #           + Hu / u^2
        out_u = -out_t * self.DPhiu
        out_u += self.zi * (D2PhiuuH - D2PhiuXH)
        out_u += self.Au * self.ui2

        lhs[:, 1] = out_u

        # ====================================================================
        # Hessian products with respect to X
        # ====================================================================
        # Hessian products for quantum entropy
        # D2_XX S(u, X) [Hx] =  Ux [log^[1](Dx) .* (Ux' Hx Ux)] Ux'
        congr_multi(self.work1, self.Ux.conj().T, self.Ax, work2)
        work1 *= self.D1x_comb
        congr_multi(work3, self.Ux, work1, work2)
        # D2_Xu S(u, X) [Hu] = -I Hu / u
        work = (self.zi * self.ui[0, 0]) * self.Au.reshape(-1, 1)
        work3[:, range(self.n), range(self.n)] -= work

        # Hessian product of barrier function
        # D2_X F(t, u, X)[Ht, Hu, Hx]
        #         = -D2_t F(t, u, X)[Ht, Hu, Hx] * D_X S(u, X)
        #           + (D2_Xu S(u, X)[Hu] + D2_XX S(u, X)[Hx]) / z
        #           + X^-1 Hx X^-1
        np.outer(out_t, self.DPhiX, out=work1.reshape((p, -1)))
        work3 -= work1

        lhs[:, 2:] = work3.reshape((p, -1)).view(np.float64)

        # Multiply A (H A')
        return dense_dot_x(lhs, A.T)

    def invhess_prod_ip(self, out, H):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.invhess_aux_updated:
            self.update_invhessprod_aux()

        (Ht, Hu, Hx) = H

        Wu = Hu + Ht * self.DPhiu
        Wx = Hx + Ht * self.DPhiX

        work = self.Ux.conj().T @ Wx @ self.Ux
        N_inv_Wx = self.Ux @ (self.D1x_comb_inv * work) @ self.Ux.conj().T
        N_inv_Wx = (N_inv_Wx + N_inv_Wx.conj().T) * 0.5

        # ======================================================================
        # Inverse Hessian products with respect to u
        # ======================================================================
        # u = (Wu + tr[N \ Wx] / z) / ((1 + tr[X]/z) / u^2 - tr[N \ I] / z^2)
        out_u = Wu * self.uz2 + np.trace(N_inv_Wx).real * self.uz
        out_u /= (self.z2 + self.trX * self.z) - self.tr_N_inv_I
        out[1][:] = out_u

        # ======================================================================
        # Inverse Hessian products with respect to X
        # ======================================================================
        # X = (N \ Wx) + u / z (N \ I)
        out_X = N_inv_Wx + (out_u / self.uz) * self.N_inv_I
        out[2][:] = out_X

        # ======================================================================
        # Inverse Hessian products with respect to t
        # ======================================================================
        # t = z^2 Ht + <DS(u, X), (u, X)>
        out_t = self.z2 * Ht + self.DPhiu * out_u + inp(self.DPhiX, out_X)
        out[0][:] = out_t

        return out

    def invhess_congr(self, A):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.invhess_aux_updated:
            self.update_invhessprod_aux()
        if not self.congr_aux_updated:
            self.congr_aux(A)

        # The inverse Hessian product applied on (Ht, Hu, Hx) for the QE barrier
        # is
        #     (u, X) =  M \ (Wu, Wx)
        #         t  =  z^2 Ht + <DPhi(u, X), (u, X)>
        # where (Wu, Wx) = [(Hu, Hx) + Ht DPhi(u, X)]
        #     M = [ (1 + tr[X]/z) / u^2   -vec(I)' / zu ]
        #         [    -vec(I) / zu              N      ]
        # and
        #     N = (Ux kron Ux) (1/z log + inv)^[1](Dx) (Ux' kron Ux')
        #
        # To solve linear systems with M, we simplify it by doing block
        # elimination, in which case we get
        #     u = (Wu + tr[N \ Wx] / z) / ((1 + tr[X]/z) / u^2 - tr[N \ I] / z^2)
        #     X = (N \ Wx) + u / z (N \ I)

        p = A.shape[0]
        lhs = np.zeros((p, sum(self.dim)))

        work1, work2, work3 = self.work1, self.work2, self.work3

        # Compute Wu
        Wu = self.Au + self.At * self.DPhiu[0, 0]
        # Compute Wx
        np.outer(self.At, self.DPhiX, out=work2.reshape((p, -1)))
        np.add(self.Ax, work2, out=work1)

        # Compute N \ Wx
        congr_multi(work2, self.Ux.conj().T, work1, work3)
        work2 *= self.D1x_comb_inv
        congr_multi(work1, self.Ux, work2, work3)

        # ======================================================================
        # Inverse Hessian products with respect to u
        # ======================================================================
        # u = (Wu + tr[N \ Wx] / z) / ((1 + tr[X]/z) / u^2 - tr[N \ I] / z^2)
        tr_N_inv_Wx = np.trace(work1, axis1=1, axis2=2).real
        out_u = Wu * self.uz2[0, 0] + tr_N_inv_Wx * self.uz[0, 0]
        out_u /= (self.z2 + self.trX * self.z) - self.tr_N_inv_I
        lhs[:, 1] = out_u

        # ======================================================================
        # Inverse Hessian products with respect to X
        # ======================================================================
        # X = (N \ Wx) + u / z (N \ I)
        np.outer(out_u / self.uz, self.N_inv_I, out=work2.reshape((p, -1)))
        work1 += work2
        out_X = work1.reshape((p, -1)).view(np.float64)
        lhs[:, 2:] = out_X

        # ======================================================================
        # Inverse Hessian products with respect to t
        # ======================================================================
        DPhiX_vec = self.DPhiX.view(np.float64).reshape((-1, 1))

        # t = z^2 Ht + <DS(u, X), (u, X)>
        out_t = self.z2 * self.At
        out_t += out_u * self.DPhiu[0, 0]
        out_t += (out_X @ DPhiX_vec).ravel()
        lhs[:, 0] = out_t

        # Multiply A (H A')
        return dense_dot_x(lhs, A.T)

    def third_dir_deriv_axpy(self, out, H, a=True):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.dder3_aux_updated:
            self.update_dder3_aux()

        (Ht, Hu, Hx) = H

        Hu2 = Hu * Hu
        trHx = np.trace(Hx).real

        chi = (Ht - self.DPhiu * Hu)[0, 0] - inp(self.DPhiX, Hx)
        chi2 = chi * chi

        UxHxUx = self.Ux.conj().T @ Hx @ self.Ux

        # Quantum entropy Hessians
        D2PhiuuH = self.trX * Hu * self.ui2
        D2PhiuXH = -np.trace(Hx).real * self.ui
        D2PhiXuH = -np.eye(self.n) * Hu * self.ui
        D2PhiXXH = self.Ux @ (self.D1x_log * UxHxUx) @ self.Ux.conj().T

        D2PhiuHH = inp(Hu, D2PhiuXH + D2PhiuuH)
        D2PhiXHH = inp(Hx, D2PhiXXH + D2PhiXuH)

        # Quantum entropy third order derivatives
        D3Phiuuu = -2 * Hu2 * self.trX * self.ui3
        D3PhiuXu = Hu * trHx * self.ui2
        D3PhiuuX = D3PhiuXu

        D3PhiXXX = scnd_frechet(self.D2x_log, UxHxUx, UxHxUx, self.Ux)
        D3PhiXuu = (Hu2 * self.ui2) * np.eye(self.n)

        # Third derivatives of barrier
        dder3_t = -2 * self.zi3 * chi2 - self.zi2 * (D2PhiXHH + D2PhiuHH)

        dder3_u = -dder3_t * self.DPhiu
        dder3_u -= 2 * self.zi2 * chi * (D2PhiuXH + D2PhiuuH)
        dder3_u += self.zi * (D3PhiuuX + D3PhiuXu + D3Phiuuu)
        dder3_u -= 2 * Hu2 * self.ui3

        dder3_X = -dder3_t * self.DPhiX
        dder3_X -= 2 * self.zi2 * chi * (D2PhiXXH + D2PhiXuH)
        dder3_X += self.zi * (D3PhiXXX + D3PhiXuu)
        dder3_X -= 2 * self.inv_X @ Hx @ self.inv_X @ Hx @ self.inv_X
        dder3_X = (dder3_X + dder3_X.conj().T) * 0.5

        out[0][:] += dder3_t * a
        out[1][:] += dder3_u * a
        out[2][:] += dder3_X * a

        return out

    # ==========================================================================
    # Auxilliary functions
    # ==========================================================================
    def congr_aux(self, A):
        assert not self.congr_aux_updated

        iscomplex = self.iscomplex

        # Get slices and views of A matrix to be used in congruence computations
        if sp.sparse.issparse(A):
            A = A.tocsr()
        self.Ax_vec = A[:, 2:]

        if sp.sparse.issparse(A):
            A = A.toarray()
        Ax_dense = np.ascontiguousarray(A[:, 2:])
        self.At = A[:, 0]
        self.Au = A[:, 1]
        self.Ax = np.array([vec_to_mat(Ax_k, iscomplex) for Ax_k in Ax_dense])

        # Preallocate matrices we will need when performing these congruences
        self.work1 = np.empty_like(self.Ax)
        self.work2 = np.empty_like(self.Ax)
        self.work3 = np.empty_like(self.Ax)

        self.congr_aux_updated = True

    def update_hessprod_aux(self):
        assert not self.hess_aux_updated
        assert self.grad_updated

        self.zi2 = self.zi * self.zi
        self.ui2 = self.ui * self.ui

        D1x_inv = np.reciprocal(np.outer(self.Dx, self.Dx))
        self.D1x_log = D1_log(self.Dx, self.log_Dx)
        self.D1x_comb = self.zi * self.D1x_log + D1x_inv

        self.hess_aux_updated = True

    def update_invhessprod_aux(self):
        assert not self.invhess_aux_updated
        assert self.grad_updated
        assert self.hess_aux_updated

        self.z2 = self.z * self.z
        self.uz = self.u * self.z
        self.uz2 = self.uz * self.uz
        self.D1x_comb_inv = np.reciprocal(self.D1x_comb)

        N_inv_I_rt2 = self.Ux * np.sqrt(np.diag(self.D1x_comb_inv))
        self.N_inv_I = N_inv_I_rt2 @ N_inv_I_rt2.conj().T
        self.tr_N_inv_I = np.trace(self.N_inv_I).real

        self.invhess_aux_updated = True

    def update_dder3_aux(self):
        assert not self.dder3_aux_updated
        assert self.hess_aux_updated

        self.zi3 = self.zi * self.zi2
        self.ui3 = self.ui * self.ui2
        self.D2x_log = D2_log(self.Dx, self.D1x_log)

        self.dder3_aux_updated = True



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import itertools
import math
import sys
import time

import numpy as np
import scipy as sp

from qics import __version__
from qics._stepper import KKTSolver, NonSymStepper, SymStepper
from qics._utils.linalg import is_full_col_rank, norm_inf
from qics.cones import OpPerspecEpi, OpPerspecTr, QuantRelEntr
from qics.point import Point

spinner = itertools.cycle(["-", "/", "|", "\\"])


class Solver:
    r"""A class representing an instance of a solver, and is parameterized
    by a conic program model and adjustable solver settings.

    Parameters
    ----------
    model : :class:`~qics.Model`
        Representation of a given conic program.
    max_iter : :obj:`int`, optional
        Maximum number of solver iterations before terminating. The default
        is ``100``.
    max_time : :obj:`float`, optional
        Maximum time elapsed, in seconds, before terminating. The default
        is ``3600``.
    tol_gap : :obj:`float`, optional
        Stopping tolerance for (relative) optimality gap. The default is
        ``1e-8``.
    tol_feas : :obj:`float`, optional
        Stopping tolerance for (relative) primal and dual feasibility. The
        default is ``1e-8``.
    tol_infeas : :obj:`float`, optional
        Tolerance for detecting infeasible problem. The default is
        ``1e-12``.
    tol_ip : :obj:`float`, optional
        Tolerance for detecting ill-posed problem. The default is
        ``1e-13``.
    tol_near : :obj:`float`, optional
        Allowable margin for certifying near optimality when solver is
        stopped early. The default is ``1e3``.
    verbose : {``0``, ``1``, ``2``, ``3``}, optional
        Verbosity level of the solver, where

        - ``0`` : No output.
        - ``1`` : Only print problem and solution summary.
        - ``2`` : Also print summary of the solver at each iteration.
        - ``3`` : Also print summary of the stepper at each iteration.

        The default is ``2``.
    ir : :obj:`bool`, optional
        Whether to use iterative refinement when solving the KKT system.
        The default is ``True``.
    toa : :obj:`bool`, optional
        Whether to use third-order adjustments to improve the stepping
        directions. The default is ``True``.
    init_pnt : :class:`~qics.point.Point`, optional
        Where to initialize the interior-point algorithm from. Variables
        which contain :obj:`~numpy.nan` are flagged to be intialized using
        the default initialization method. The default is ``None``, which
        intializes all variables using the default method.
    use_invhess : :obj:`bool`, optional
        Whether or not to avoid using inverse Hessian product oracles by
        solving a modified cone program with
        :math:`G^{-1}(\mathcal{K})=\{ x : Gx \in \mathcal{K} \}`. By 
        default, this is ``False`` if :math:`G\neq-\mathbb{I}`, :math:`G`
        is full column rank, and :math:`\mathcal{K}` mainly consists of
        :class:`~qics.cones.QuantRelEntr`, 
        :class:`~qics.cones.OpPerspecEpi`, and
        :class:`~qics.cones.OpPerspecTr` cones, and is ``True`` otherwise.
    """

    def __init__(
        self,
        model,
        max_iter=100,
        max_time=3600,
        tol_gap=1e-8,
        tol_feas=1e-8,
        tol_infeas=1e-12,
        tol_ip=1e-13,
        tol_near=1e3,
        verbose=2,
        ir=True,
        toa=True,
        init_pnt=None,
        use_invhess=None,
    ):
        # Basic solver options
        self.max_iter = max_iter
        self.max_time = max_time
        self.verbose = verbose

        self.tol_gap = tol_gap
        self.tol_feas = tol_feas
        self.tol_infeas = tol_infeas
        self.tol_ip = tol_ip
        self.tol_near = tol_near

        # Slow progress options
        self.small_step_tol = 0.005
        self.consecutive_small_step_limit = 2
        self.consecutive_small_steps = 0

        # Avoiding inverse Hessian options
        cones = model.cones
        if use_invhess is None:
            # Default decision tree for whether to avoid inverse Hessian oracles
            SLOW_CONES = (QuantRelEntr, OpPerspecEpi, OpPerspecTr)
            q_slow_cones = sum([sum(K.dim) for K in cones if isinstance(K, SLOW_CONES)])

            use_invhess = model.issymmetric or q_slow_cones / model.q <= 0.55 \
                          or not model.use_G or not is_full_col_rank(model.G)  # fmt: skip
        elif not use_invhess:
            if not model.use_G:
                raise Exception(
                    "Avoiding inverse Hessian oracles is not supported nor recommended "
                    "if G is easily invertible."
                )
            if not is_full_col_rank(model.G):
                raise Exception(
                    "Avoiding inverse Hessian oracles is not supported when G is not "
                    "full column rank."
                )
        self.use_invhess = use_invhess

        # Preprocess model
        self.model = model
        model._preprocess(use_invhess, init_pnt)

        # Initialize steppers
        kktsolver = KKTSolver(model, ir=ir, use_invhess=use_invhess)
        if model.issymmetric:
            self.stepper = SymStepper(kktsolver, model, toa=toa)
        else:
            self.stepper = NonSymStepper(kktsolver, model, toa=toa)

        # Initialize points and process initial point
        self.point = Point(model)
        self.point_best = Point(model)

        self.point.vec.fill(np.nan)
        if init_pnt is not None:
            n_orig, p_orig = model.n_orig, model.p_orig
            self.point.x[:n_orig] = init_pnt.x * model.c_scale[:n_orig]
            self.point.y[:p_orig] = init_pnt.y * model.b_scale[:p_orig]
            self.point.z.vec[:] = init_pnt.z.vec * model.h_scale
            self.point.s.vec[:] = init_pnt.s.vec * model.h_scale
            self.point.tau[:] = init_pnt.tau
            self.point.kap[:] = init_pnt.kap

        return

    def solve(self):
        r"""Run the primal-dual interior point solver for a given problem
        model.

        Returns
        -------
        :obj:`dict`
            Dictionary which summarizes the solution of a conic program.
            Contains the following keys.

            - x_opt (:obj:`~numpy.ndarray`) : Optimal primal variable
              :math:`x^*`.
            - y_opt (:obj:`~numpy.ndarray`) : Optimal dual variable
              :math:`y^*`.
            - z_opt (:obj:`~qics.point.VecProduct`) : Optimal dual variable
              :math:`z^*`.
            - s_opt (:obj:`~qics.point.VecProduct`) : Optimal primal
              variable :math:`s^*=h-Gx^*`.
            - sol_status (:obj:`string`) : Solution status. Can either be:

                - ``optimal``      : Primal-dual optimal solution reached
                - ``pinfeas``      : Detected primal infeasibility
                - ``dinfeas``      : Detected dual infeasibility
                - ``near_optimal`` : Near primal-dual optimal solution
                - ``near_pinfeas`` : Near primal infeasibility
                - ``near_dinfeas`` : Near dual infeasibiltiy
                - ``illposed``     : Problem is ill-posed
                - ``unknown``      : Unknown solution status

            - exit_status (:obj:`string`) : Solver exit status. Can either
              be:

                - ``solved``       : Terminated at desired tolerance
                - ``max_iter``     : Exceeded maximum allowable iterations
                - ``max_time``     : Exceeded maximum allowable time
                - ``step_failure`` : Unable to take another step
                - ``slow_progress``: Residuals are decreasing too slowly

            - num_iter (:obj:`int`) : Number of solver iterations.
            - solve_time (:obj:`float`) : Total time elapsed (in seconds).
            - p_obj (:obj:`float`) : Optimal primal objective
              :math:`c^\top x^*`.
            - d_obj (:obj:`float`) : Optimal dual objective
              :math:`-b^\top y^* - h^\top z^*`.
            - opt_gap (:obj:`float`) : Relative optimality gap.
            - p_feas (:obj:`float`) : Relative primal feasibility.
            - d_feas (:obj:`float`) : Relative dual feasibiltiy.
        """
        # Setup solver
        self.setup_solver()
        self.setup_point()
        self.calc_mu()
        self.get_gap_feas()

        # Print header
        if self.verbose:
            self.print_title()
            if self.verbose >= 2:
                self.print_iter_heading()
                self.print_iter()
            else:
                print()
                sys.stdout.write("Running...")

        # Run main solver loop
        while True:
            if self.step_and_check():
                break

        # Solver has termianted, perform final clean up
        self.solve_time = time.time() - self.start_time

        # If we didn't reach a solution, check if we are close to optimal
        if self.exit_status != "solved":
            self.copy_solver_data()
            self.retrieve_best_data()

            self.solution_status = "unknown"

            # Check near optimality
            if self.gap <= self.tol_gap * self.tol_near:
                if self.x_feas <= self.tol_feas * self.tol_near:
                    if self.y_feas <= self.tol_feas * self.tol_near:
                        if self.z_feas <= self.tol_feas * self.tol_near:
                            self.solution_status = "near_optimal"

            # Check near infeasibility
            if self.x_infeas <= self.tol_infeas * self.tol_near:
                self.solution_status = "near_pinfeas"

            if self.y_infeas <= self.tol_infeas * self.tol_near:
                if self.z_infeas <= self.tol_infeas * self.tol_near:
                    self.solution_status = "near_dinfeas"

        if self.verbose:
            self.print_solution()

        # Scale variables back
        x_opt = self.point.x / self.model.c_scale / self.point.tau
        y_opt = self.point.y / self.model.b_scale / self.point.tau
        if not self.use_invhess:
            x_opt = (x_opt + self.model.x_offset)[: self.model.n_orig]
            y_opt = y_opt[: self.model.p_orig]
        z_opt = self.point.z
        z_opt.vec[:] = z_opt.vec / self.model.h_scale / self.point.tau
        s_opt = self.point.s
        s_opt.vec[:] = s_opt.vec / self.model.h_scale / self.point.tau

        return {
            "x_opt": x_opt,
            "y_opt": y_opt,
            "z_opt": z_opt,
            "s_opt": s_opt,
            "sol_status": self.solution_status,
            "exit_status": self.exit_status,
            "num_iter": self.iter,
            "solve_time": self.solve_time,
            "p_obj": self.p_obj,
            "d_obj": self.d_obj,
            "opt_gap": self.gap,
            "p_feas": max(self.y_feas, self.z_feas),
            "d_feas": self.x_feas,
        }

    def step_and_check(self):
        # ==============================================================
        # Step
        # ==============================================================
        # Make a copy of current point before taking a step
        self.copy_solver_data()

        # Take a step
        self.point, success, alpha = self.stepper.step(
            self.model, self.point, self.res, self.mu, self.verbose
        )
        self.iter += 1
        self.elapsed_time = time.time() - self.start_time

        # Compute barrier parameter and residuals
        self.calc_mu()
        self.get_gap_feas()

        # ==============================================================
        # Print iteration status
        # ==============================================================
        if self.verbose >= 2:
            self.print_iter()
        elif self.verbose:
            sys.stdout.write(next(spinner))
            sys.stdout.flush()
            sys.stdout.write("\b")

        # ==============================================================
        # Check termination criteria
        # ==============================================================
        # 1) Check optimality
        if self.gap <= self.tol_gap:
            if (
                self.x_feas <= self.tol_feas
                and self.y_feas <= self.tol_feas
                and self.z_feas <= self.tol_feas
            ):
                self.solution_status = "optimal"
                self.exit_status = "solved"
                return True

        # 2) Check primal and dual infeasibility
        if self.x_infeas <= self.tol_infeas:
            self.solution_status = "pinfeas"
            self.exit_status = "solved"
            return True

        if self.y_infeas <= self.tol_infeas and self.z_infeas <= self.tol_infeas:
            self.solution_status = "dinfeas"
            self.exit_status = "solved"
            return True

        # 3) Check ill-posedness
        if self.point.tau < self.tol_infeas and self.point.kap < self.tol_infeas:
            if self.illposed_res <= self.tol_ip:
                self.solution_status = "illposed"
                self.exit_status = "solved"
                return True

        # 4) Check if maximum iterations is exceeded
        if self.iter >= self.max_iter:
            self.exit_status = "max_iter"
            return True

        # 5) Check if maximum time is exceeded
        if self.elapsed_time >= self.max_time:
            self.exit_status = "max_time"
            return True

        # 6) Did the step fail or not
        if not success:
            self.exit_status = "step_failure"
            return True

        # 7) Check if progress is slow or degrading at high tolerance
        if alpha <= self.small_step_tol:
            self.consecutive_small_steps += 1
            if self.consecutive_small_steps >= self.consecutive_small_step_limit:
                self.exit_status = "slow_progress"
                return True
        else:
            self.consecutive_small_steps = 0

        return False

    def setup_solver(self):
        self.iter = 0
        self.start_time = time.time()
        self.elapsed_time = 0.0

        model = self.model

        self.c_max = norm_inf(model.c)
        self.b_max = norm_inf(model.b)
        self.h_max = norm_inf(model.h)

        if self.verbose == 3:
            if self.model.issymmetric:
                self.printbar_size = 125
            else:
                self.printbar_size = 136
        else:
            self.printbar_size = 97

        return

    def setup_point(self):
        model = self.model

        if any(np.isnan(self.point.x)):
            self.point.x.fill(0.0)
        if any(np.isnan(self.point.y)):
            self.point.y.fill(0.0)

        for k, cone_k in enumerate(model.cones):
            if any(np.isnan(self.point.s.vecs[k])):
                cone_k.get_init_point(self.point.s[k])
            else:
                cone_k.set_point(self.point.s[k])
            if not cone_k.get_feas():
                raise ValueError("Initial primal variable s is infeasible.")

            if any(np.isnan(self.point.z.vecs[k])):
                cone_k.grad_ip(self.point.z[k])
                self.point.z.vecs[k] *= -1
            cone_k.set_dual(self.point.z[k])
            if not cone_k.get_dual_feas():
                raise ValueError("Initial dual variable z is infeasible.")

        if any(np.isnan(self.point.tau)):
            self.point.tau.fill(1.0)
        if any(np.isnan(self.point.kap)):
            self.point.kap.fill(1.0)

        self.calc_mu()
        if not math.isclose(self.mu, 1.0):
            print(f"Initial mu is {self.mu} but should be 1")

        return

    def calc_mu(self):
        s_inp_z = self.point.s.inp(self.point.z)
        kap_inp_tau = self.point.tau[0, 0] * self.point.kap[0, 0]
        self.mu = (s_inp_z + kap_inp_tau) / self.model.nu
        return self.mu

    def get_gap_feas(self):
        model = self.model
        c = self.model.c
        b = self.model.b
        h = self.model.h

        x = self.point.x
        y = self.point.y
        z = self.point.z.vec
        s = self.point.s.vec
        tau = self.point.tau[0, 0]
        kap = self.point.kap[0, 0]

        self.kap_tau = kap / tau

        # Get primal and dual objectives and optimality gap
        p_obj_tau = (c.T @ x)[0, 0]
        d_obj_tau = -(b.T @ y + h.T @ z)[0, 0]

        self.p_obj = p_obj_tau / tau + self.model.offset
        self.d_obj = d_obj_tau / tau + self.model.offset

        gap1 = self.point.z.inp(self.point.s) / tau
        gap2 = abs(p_obj_tau - d_obj_tau)
        self.gap = min(gap1, gap2)
        self.gap /= max(tau, min([abs(p_obj_tau), abs(d_obj_tau)]))

        # Get primal and dual infeasibilities
        self.x_res = model.A_T @ y
        self.x_res += model.G_T @ z
        self.x_res *= -1
        self.y_res = model.A @ x
        self.z_res = model.G @ x
        self.z_res += s

        norm_x_res = norm_inf(self.x_res)
        norm_y_res = norm_inf(self.y_res)
        norm_z_res = norm_inf(self.z_res)

        self.x_infeas = norm_x_res / d_obj_tau if (d_obj_tau > 0) else np.inf
        self.y_infeas = -norm_y_res / p_obj_tau if (p_obj_tau < 0) else np.inf
        self.z_infeas = -norm_z_res / p_obj_tau if (p_obj_tau < 0) else np.inf

        # Get ill posedness certificates
        norm_xyzs = max(norm_inf(x), norm_inf(y), norm_inf(z), norm_inf(s))
        if norm_xyzs > 0:
            norm_xyz_res = max(norm_x_res, norm_y_res, norm_z_res)
            self.illposed_res = norm_xyz_res / norm_xyzs
        else:
            self.illposed_res = np.inf

        # Get primal and dual feasibilities
        self.x_res = sp.linalg.blas.daxpy(c, self.x_res, a=-tau)
        if model.p > 0:
            self.y_res = sp.linalg.blas.daxpy(b, self.y_res, a=-tau)
        if model.q > 0:
            self.z_res = sp.linalg.blas.daxpy(h, self.z_res, a=-tau)
        self.tau_res = p_obj_tau - d_obj_tau + kap

        self.x_feas = norm_inf(self.x_res) / (1.0 + self.c_max) / tau
        self.y_feas = norm_inf(self.y_res) / (1.0 + self.b_max) / tau
        self.z_feas = norm_inf(self.z_res) / (1.0 + self.h_max) / tau

        self.res = {
            "x": self.x_res,
            "y": self.y_res,
            "z": self.z_res,
            "tau": self.tau_res,
        }

        return

    def copy_solver_data(self):
        gap_feas = max(self.x_feas, self.y_feas, self.z_feas, self.gap)
        if self.iter == 0:
            # Best point hasn't been initialized yet
            gap_feas_best = np.inf
        else:
            gap_feas_best = max(self.x_feas_best, self.y_feas_best,
                                self.z_feas_best, self.gap_best)  # fmt: skip

        if gap_feas_best > gap_feas:
            self.point_best.vec[:] = self.point.vec
            self.best_iter = self.iter

            self.p_obj_best = self.p_obj
            self.d_obj_best = self.d_obj
            self.gap_best = self.gap

            self.x_feas_best = self.x_feas
            self.y_feas_best = self.y_feas
            self.z_feas_best = self.z_feas
            self.tau_res_best = self.tau_res

            self.x_infeas_best = self.x_infeas
            self.y_infeas_best = self.y_infeas
            self.z_infeas_best = self.z_infeas

    def retrieve_best_data(self):
        if self.best_iter != self.iter:
            if self.verbose:
                print("\nRetrieving data from iteration ", self.best_iter)

            self.point.vec = self.point_best.vec[:]

            self.p_obj = self.p_obj_best
            self.d_obj = self.d_obj_best
            self.gap = self.gap_best

            self.x_feas = self.x_feas_best
            self.y_feas = self.y_feas_best
            self.z_feas = self.z_feas_best
            self.tau_res = self.tau_res_best

            self.x_infeas = self.x_infeas_best
            self.y_infeas = self.y_infeas_best
            self.z_infeas = self.z_infeas_best

    def print_title(self):
        filler = "="
        print(f"{'':{filler}^{68}}")
        print(f"{'QICS v' + __version__ + ' - Quantum Information Conic Solver':^68}")
        print(f"{'by K. He, J. Saunderson, H. Fawzi (2024)':^68}")
        print(f"{'':{filler}^{68}}")

        print("Problem summary:")
        print(f"\tno. vars:     {self.model.n:<10}", end="")
        print(f"\t\tbarr. par:    {self.model.nu:<10}")
        print(f"\tno. constr:   {self.model.p:<10}", end="")
        print(f"\t\tsymmetric:    {self.model.issymmetric!r:<10}")
        print(f"\tcone dim:     {self.model.q:<10}", end="")
        print(f"\t\tcomplex:      {self.model.iscomplex!r:<10}")
        print(f"\tno. cones:    {len(self.model.cones):<10}", end="")
        print(f"\t\tsparse:       {self.model.issparse!r:<10}")

    def print_iter_heading(self):
        filler = "="
        print(f"\n{'':{filler}^{self.printbar_size}}", end="")
        print(f"\n {'iter':^4}   {'mu':^7}   {'k/t':^7}  ", end="")
        print(f"|  {'p_obj':^10}  {'d_obj':^10}  {'gap':^7}  ", end="")
        print(f"|  {'p_feas':^7}   {'d_feas':^7}  ", end="")
        print(f"|  {'time (s)':^8}  ", end="")
        if self.verbose == 3:
            if self.model.issymmetric:
                print(f"|  {'dir_tol':^7}   {'sigma':^5}   {'alpha':^5}", end="")
            else:
                print(f"|  {'step':^6}   {'dir_tol':^7}   ", end="")
                print(f"{'prox':^7}   {'alpha':^5}", end="")
        print(f"\n{'':{filler}^{self.printbar_size}}", end="")

    def print_iter(self):
        yz_feas = max(self.y_feas, self.z_feas)
        print(f"\n {self.iter:>4}   {self.mu:>7.1e}   {self.kap_tau:>7.1e}  ", end="")
        print(f"| {self.p_obj:>10.3e}  {self.d_obj:>10.3e}  {self.gap:>8.1e}  ", end="")
        print(f"|  {yz_feas:>7.1e}   {self.x_feas:>7.1e}  ", end="")
        print(f"|  {self.elapsed_time:<8.2f}", end="")

    def print_solution(self):
        print("\n\nSolution summary")
        print(f"\tsol. status:  {self.solution_status:<19}", end="")
        print(f"\tnum. iter:    {self.iter:<10}")
        print(f"\texit status:  {self.exit_status:<19}", end="")
        print(f"\tsolve time:   {self.solve_time:<10.3f}")
        print(f"\tprimal obj:  {self.p_obj:>19.12e}", end="")
        print(f"\tprimal feas:  {max(self.y_feas, self.z_feas):<.2e}")
        print(f"\tdual obj:    {self.d_obj:>19.12e}", end="")
        print(f"\tdual feas:    {self.x_feas:<.2e}")
        print()



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

# __init__.py
from qics._version import __version__  # noqa

from qics.solver import Solver  # noqa
from qics.model import Model  # noqa

import qics.io  # noqa



# setup.py
import os
import re

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

PKG = "qics"
VERSIONFILE = os.path.join(PKG, "_version.py")
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    VERSION = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

setuptools.setup(
    name="qics",
    version=VERSION,
    author="Kerry He, James Saunderson, and Hamza Fawzi",
    author_email="he.kerry.k@gmail.com, james.saunderson@monash.edu, h.fawzi@damtp.cam.ac.uk",
    description="Conic solver for quantum information theory",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    project_urls={
        "Source": "https://github.com/kerry-he/qics",
        "Documentation": "https://qics.readthedocs.io/en/latest/",
    },
    packages=setuptools.find_packages(include=["qics", "qics.*"]),
    python_requires=">=3.8",
    install_requires=["numpy", "scipy", "numba"],
    package_data={"": ["README.md", "LICENSE.md"]},
)



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi
# Based on test_examples.py from CVXOPT by M. Andersen and L. Vandenberghe.

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

def test_ptr_ikr():
    # Tests that p_tr and i_kr satisfy the adjoint relationship
    #     <p_tr(X), Y> = <X, i_kr(Y)>
    import numpy as np

    from qics.quantum import i_kr, p_tr

    np.random.seed(1)

    dims = [2, 4, 3, 5]

    x_dim = np.prod(dims)
    X = np.random.randn(x_dim, x_dim) + np.random.randn(x_dim, x_dim) * 1j
    X = X + X.conj().T

    sys_tests = [
        0,
        1,
        2,
        3,
        (0, 1),
        (2, 0),
        (0, 3),
        (2, 1),
        (1, 3),
        (3, 2),
        (0, 1, 2),
        (3, 0, 1),
        (0, 3, 2),
        (3, 2, 1),
        (0, 1, 2, 3),
    ]

    for sys in sys_tests:
        dim = dims[sys] if isinstance(sys, int) else np.prod([dims[k] for k in sys])
        y_dim = x_dim // dim
        Y = np.random.randn(y_dim, y_dim) + np.random.randn(y_dim, y_dim) * 1j
        Y = Y + Y.conj().T

        assert np.allclose(
            np.trace(p_tr(X, dims, sys) @ Y), np.trace(X @ i_kr(Y, dims, sys))
        ), "qics.quantum.p_tr and qics.quantum.i_kr are not adjoint linear operators"


def test_ptr_ikr_multi():
    # Tests that p_tr_multi and i_kr_multi satisfy the adjoint relationship
    #     <p_tr_multi(Xs), Ys> = <Xs, i_kr_multi(Ys)>
    import numpy as np

    from qics.quantum import i_kr_multi, p_tr_multi

    np.random.seed(1)

    dims = [2, 4, 3, 5]

    x_dim = np.prod(dims)
    Xs = np.random.randn(5, x_dim, x_dim) + np.random.randn(5, x_dim, x_dim) * 1j
    Xs = Xs + Xs.conj().transpose(0, 2, 1)

    sys_tests = [
        0,
        1,
        2,
        3,
        (0, 1),
        (2, 0),
        (0, 3),
        (2, 1),
        (1, 3),
        (3, 2),
        (0, 1, 2),
        (3, 0, 1),
        (0, 3, 2),
        (3, 2, 1),
        (0, 1, 2, 3),
    ]

    for sys in sys_tests:
        dim = dims[sys] if isinstance(sys, int) else np.prod([dims[k] for k in sys])
        y_dim = x_dim // dim
        Ys = np.random.randn(5, y_dim, y_dim) + np.random.randn(5, y_dim, y_dim) * 1j
        Ys = Ys + Ys.conj().transpose(0, 2, 1)

        assert np.allclose(
            np.trace(
                p_tr_multi(np.ones_like(Ys), Xs, dims, sys) @ Ys, axis1=1, axis2=2
            ),
            np.trace(
                Xs @ i_kr_multi(np.ones_like(Xs), Ys, dims, sys), axis1=1, axis2=2
            ),
        ), "qics.quantum.p_tr_multi and qics.quantum.i_kr_multi are not adjoint linear operators"



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import numpy as np
import scipy as sp

from qics._utils.gradient import D1_log, D2_log, scnd_frechet
from qics._utils.linalg import (
    cho_fact,
    cho_solve,
    congr_multi,
    dense_dot_x,
    inp,
    x_dot_dense,
)
from qics.cones.base import Cone
from qics.vectorize import get_full_to_compact_op, vec_to_mat


class QuantKeyDist(Cone):
    r"""A class representing a quantum key distribution cone

    .. math::

        \mathcal{QKD}_{\mathcal{G},\mathcal{Z}} =
        \text{cl}\{ (t, X) \in \mathbb{R} \times \mathbb{H}^n_{++} :
        t \geq -S(\mathcal{G}(X)) +  S(\mathcal{Z}(\mathcal{G}(X))) \},

    where

    .. math::

        S(X) = -\text{tr}[X \log(X)],

    is the quantum (von Neumann) entropy function,
    :math:`\mathcal{G}:\mathbb{H}^n\rightarrow\mathbb{H}^{mr}` is a
    positive linear map, and
    :math:`\mathcal{Z}:\mathbb{H}^{mr}\rightarrow\mathbb{H}^{mr}` is a
    pinching map that maps off-diagonal blocks to zero.

    Parameters
    ----------
    G_info : :obj:`int` or :obj:`list` of :class:`~numpy.ndarray`
        Defines the linear map :math:`\mathcal{G}`. There are two ways to
        specify this linear map.

        - If ``G_info`` is an :obj:`int`, then :math:`\mathcal{G}` is the
          identity map, i.e., :math:`\mathcal{G}(X)=X`, and ``G_info``
          specifies the dimension of :math:`X`.
        - If ``G_info`` is a :obj:`list` of :class:`~numpy.ndarray`, then
          ``G_info`` specifies the Kraus operators
          :math:`K_i \in \mathbb{C}^{mr \times n }` corresponding to
          :math:`\mathcal{G}` such that

          .. math::

              \mathcal{G}(X) = \sum_{i} K_i X K_i^\dagger.

    Z_info : :obj:`int` or :obj:`tuple` or :obj:`list` of :class:`~numpy.ndarray`

        Defines the pinching map :math:`\mathcal{Z}`, which is of the form

        .. math::

            \mathcal{Z}(Y) = \sum_{i} Z_i Y Z_i^\dagger.

        There are three ways to specify this linear map.

        - If ``Z_info`` is an :obj:`int`, then
          :math:`Z_i=|i \rangle\langle i| \otimes\mathbb{I}` for
          :math:`i=1,\ldots,r`, where ``r=Z_info``.

        - If ``Z_info`` is a :obj:`tuple` of the form ``(dims, sys)``,
          where ``dims=(n0, n1)`` is a :obj:`tuple` of :obj:`int` and
          ``sys`` is an :obj:`int`, then

          - :math:`Z_i=|i \rangle\langle i| \otimes\mathbb{I}_{n_1}`
            for :math:`i=1,\ldots,n_0` if ``sys=0``, and
          - :math:`Z_i=\mathbb{I}_{n_0}\otimes |i \rangle\langle i|` for
            :math:`i=1,\ldots,n_1` if ``sys=1``.

          We generalize this definition to the case where ``dims`` and ``sys``
          are lists of any length.

        - If ``Z_info`` is a :obj:`list` of :class:`~numpy.ndarray`, then
          ``Z_info`` directly specifies the Kraus operators
          :math:`Z_i \in \mathbb{C}^{mr \times mr}`. Note that these Kraus
          operators must have a similar structure to those defined using
          the other options, i.e., must be diagonal matrices consisting of
          either ones or zeros, and :math:`Z_iZ_j=0` for all :math:`i\neq j`.

    iscomplex : :obj:`bool`
        Whether the matrix :math:`X` is defined over :math:`\mathbb{H}^n`
        (``True``), or restricted to :math:`\mathbb{S}^n` (``False``). The
        default is ``False``.

    See also
    --------
    QuantRelEntr : Quantum relative entropy cone

    Notes
    -----
    The quantum key distribution cone can also be modelled by the quantum
    relative entropy by noting the identity

    .. math::

        S(\mathcal{G}(X) \| \mathcal{Z}(\mathcal{G}(X)))
        = -S(\mathcal{G}(X)) +  S(\mathcal{Z}(\mathcal{G}(X))).

    However, the cone oracles for the quantum key distribution cone are
    more efficient than those for the quantum relative entropy cone
    (especially when :math:`\mathcal{G}` is the idenity map), so it is
    recommended to use the quantum key distribution cone where possible.
    """

    def __init__(self, G_info, Z_info, iscomplex=False):
        self._process_G_info(G_info)
        self._process_Z_info(Z_info)
        self.iscomplex = iscomplex

        self.nu = 1 + self.n  # Barrier parameter

        if iscomplex:
            self.vn = self.n * self.n
            self.vm = self.m * self.m
            self.dim = [1, 2 * self.n * self.n]
            self.type = ["r", "h"]
            self.dtype = np.complex128
        else:
            self.vn = self.n * (self.n + 1) // 2
            self.vm = self.m * (self.m + 1) // 2
            self.dim = [1, self.n * self.n]
            self.type = ["r", "s"]
            self.dtype = np.float64

        # Facial reduction on G(X) and Z(G(X))
        ZK_list_raw = [[K[Z, :] for K in self.K_list_raw] for Z in self.Z_idxs]
        self.K_list_blk = [facial_reduction(self.K_list_raw)]
        self.ZK_list_blk = [facial_reduction(ZK) for ZK in ZK_list_raw]

        self.Nk = [K_list[0].shape[0] for K_list in self.K_list_blk]
        self.Nzk = [ZK_list[0].shape[0] for ZK_list in self.ZK_list_blk]

        # Update flags
        self.feas_updated = False
        self.grad_updated = False
        self.hess_aux_updated = False
        self.invhess_aux_updated = False
        self.dder3_aux_updated = False
        self.invhess_aux_aux_updated = False
        self.congr_aux_updated = False
        self.hess_congr_aux_updated = False
        self.invhess_congr_aux_updated = False

        if self.G_is_Id:
            self.precompute_computational_basis(self.m)
            self.F2C_op = get_full_to_compact_op(self.m, iscomplex)
        else:
            self.precompute_computational_basis(self.n)
            self.F2C_op = get_full_to_compact_op(self.n, iscomplex)

        return

    def _process_G_info(self, G_info):
        if isinstance(G_info, int):
            # Define G(X) as the identity map
            self.n = n = G_info  # Input dimension of X
            self.N = N = G_info  # Output dimension of G(X)

            self.K_list_raw = [np.eye(self.N)]
            self.G_is_Id = True

        else:
            # Define G(X) using given Kraus operators
            self.n = n = G_info[0].shape[1]  # Input dimension of X
            self.N = N = G_info[0].shape[0]  # Output dimension of G(X)
            assert all([Ki.shape == (N, n) for Ki in G_info]), "Kraus " \
                "operators specified by G_info must have the same dimensions."  # fmt: skip

            self.K_list_raw = G_info
            self.G_is_Id = len(G_info) == 1 and n == N \
                           and np.linalg.norm(G_info[0] - np.eye(n)) < 1e-10  # fmt: skip

    def _process_Z_info(self, Z_info):
        N = self.N

        if isinstance(Z_info, int):
            # Define block structure for Z(X) with r x r blocks of size m x m
            self.r = r = Z_info
            self.m = m = N // r
            assert m * r == N, "Number of blocks r specified by Z_info " \
                "must be an integer factor of the dimension of G(X)."  # fmt: skip

            self.Z_list_raw = [np.zeros((N, N)) for _ in range(r)]
            for k in range(r):
                range_k = np.arange(k * m, (k + 1) * m)
                self.Z_list_raw[k][range_k, range_k] = 1.0
            self.Z_idxs = [np.arange(i * m, (i + 1) * m) for i in range(r)]

        elif isinstance(Z_info, tuple):
            # Define custom block structure for a given subsystem structure
            (dims, sys) = Z_info
            if isinstance(sys, int):
                sys = [sys]
            if isinstance(sys, tuple) or isinstance(sys, set):
                sys = list(sys)
            assert np.prod(dims) == N, "Total dimension of subsystems must equal " \
                "the dimension of G(X)."  # fmt: skip
            assert all([sys_k < len(dims) for sys_k in sys]), "Invalid subsystems " \
                "specified, exceeds total number of dimensions provided."  # fmt: skip

            self.r = np.prod([dims[k] for k in sys])
            self.m = N // self.r

            idxs = np.meshgrid(*[range(dims[k]) for k in sys])
            idxs = list(np.array(idxs).reshape(len(sys), -1).T)
            self.Z_list_raw = []
            for i in range(len(idxs)):
                Z_i = np.array([1])
                counter = 0
                for k, dim_k in enumerate(dims):
                    if k in sys:
                        Z_ik = np.zeros(dim_k)
                        Z_ik[idxs[i][counter]] = 1
                        Z_i = np.kron(Z_i, Z_ik)
                        counter += 1
                    else:
                        Z_i = np.kron(Z_i, np.ones(dim_k))
                self.Z_list_raw += [np.diag(Z_i)]
            self.Z_idxs = [np.where(Z)[0] for Z in self.Z_list_raw]
        else:
            # Define Z(X) using given Kraus operators
            self.r = len(Z_info)
            self.m = self.N // self.r
            assert all([Zi.shape == (N, N) for Zi in Z_info]), "Kraus operators "     \
                "specified by Z_info must have the same dimensions."  # fmt: skip
            assert not any([np.any(Zi - np.diag(np.diag(Zi))) for Zi in Z_info]),     \
                "Kraus operators specified by Z_info must be diagonal."  # fmt: skip
            assert all([all([x in {0, 1} for x in np.diag(Zi)]) for Zi in Z_info]),   \
                "Diagonal elements of Kraus operators specified by Z_info must only " \
                "be either 0s or 1s."  # fmt: skip
            assert np.all(sum(Z_info) == np.eye(N)), "Kraus operators specified by "  \
                "Z_info must satisfy ZiZj = 0 for all i =/= j."  # fmt: skip

            self.Z_list_raw = Z_info
            self.Z_idxs = [np.where(Z)[0] for Z in Z_info]

    def get_iscomplex(self):
        return self.iscomplex

    def get_init_point(self, out):
        from qics.quantum import entropy

        eye = np.eye(self.n)
        GI_blk = [apply_kraus(eye, K_list) for K_list in self.K_list_blk]
        ZGI_blk = [apply_kraus(eye, ZK_list) for ZK_list in self.ZK_list_blk]

        entr_GI = sum([entropy(KK) for KK in GI_blk])
        entr_ZGI = sum([entropy(ZKKZ) for ZKKZ in ZGI_blk])

        f0 = -entr_GI + entr_ZGI
        t0 = f0 / 2 + np.sqrt(1 + f0 * f0 / 4)

        point = [np.array([[t0]]), np.eye(self.n, dtype=self.dtype)]

        self.set_point(point, point)

        out[0][:] = point[0]
        out[1][:] = point[1]

        return out

    def get_feas(self):
        if self.feas_updated:
            return self.feas

        self.feas_updated = True

        (self.t, self.X) = (t, X) = self.primal

        # Check that X is positive definite
        self.Dx, self.Ux = np.linalg.eigh(X)
        if any(self.Dx <= 0):
            self.feas = False
            return self.feas

        # Check that G(X) is positive definite
        #   Note that G(X) should be positive definite if X is, but check
        #   just to be sure that we can safely take logarithms of G(X).
        GX_blk = [apply_kraus(X, K_list) for K_list in self.K_list_blk]

        DUgx_blk = [np.linalg.eigh(GX) for GX in GX_blk]
        self.Dgx_blk = Dgx_blk = [DUgx[0] for DUgx in DUgx_blk]
        self.Ugx_blk = [DUgx[1] for DUgx in DUgx_blk]

        if any([any(Dgx <= 0) for Dgx in self.Dgx_blk]):
            self.feas = False
            return self.feas

        # Check that Z(G(X)) is positive definite
        #   Note that Z(G(X)) should be positive definite if X is, but check
        #   just to be sure that we can safely take logarithms of Z(G(X)).
        ZGX_blk = [apply_kraus(X, ZG_list) for ZG_list in self.ZK_list_blk]

        DUzgx_blk = [np.linalg.eigh(ZGX) for ZGX in ZGX_blk]
        self.Dzgx_blk = Dzgx_blk = [DUzkx[0] for DUzkx in DUzgx_blk]
        self.Uzgx_blk = [DUzkx[1] for DUzkx in DUzgx_blk]

        if any([any(Dzkx <= 0) for Dzkx in self.Dzgx_blk]):
            self.feas = False
            return self.feas

        # Check that t > -S(G(X)) + S(Z(G(X)))
        self.log_Dgx_blk = log_Dgx_blk = [np.log(D) for D in self.Dgx_blk]
        self.log_Dzgx_blk = log_Dzgx_blk = [np.log(D) for D in self.Dzgx_blk]

        entr_GX = [inp(D, log_D) for (D, log_D) in zip(Dgx_blk, log_Dgx_blk)]
        entr_ZGX = [inp(D, log_D) for (D, log_D) in zip(Dzgx_blk, log_Dzgx_blk)]
        self.z = t[0, 0] - (sum(entr_GX) - sum(entr_ZGX))

        self.feas = self.z > 0
        return self.feas

    def get_val(self):
        assert self.feas_updated

        return -np.log(self.z) - np.sum(np.log(self.Dx))

    def update_grad(self):
        assert self.feas_updated
        assert not self.grad_updated

        # Compute gradients of quantum relative entropy
        # DPhi(X) = G'[log(G(X)) + I] - G'Z'[log(G(Z(X))) + I]
        log_GX_blk = [
            (U * logD) @ U.conj().T + np.eye(n)
            for (U, logD, n) in zip(self.Ugx_blk, self.log_Dgx_blk, self.Nk)
        ]
        log_ZGX_blk = [
            (U * logD) @ U.conj().T + np.eye(n)
            for (U, logD, n) in zip(self.Uzgx_blk, self.log_Dzgx_blk, self.Nzk)
        ]

        G_log_GX = sum(
            [
                apply_kraus(logX, K_list, adjoint=True)
                for (logX, K_list) in zip(log_GX_blk, self.K_list_blk)
            ]
        )
        ZG_log_ZGX = sum(
            [
                apply_kraus(logX, ZK_list, adjoint=True)
                for (logX, ZK_list) in zip(log_ZGX_blk, self.ZK_list_blk)
            ]
        )

        self.DPhi = G_log_GX - ZG_log_ZGX

        # Compute X^-1
        inv_Dx = np.reciprocal(self.Dx)
        inv_X_rt2 = self.Ux * np.sqrt(inv_Dx)
        self.inv_X = inv_X_rt2 @ inv_X_rt2.conj().T

        # Compute gradient of barrier function
        self.zi = np.reciprocal(self.z)
        self.grad = [-self.zi, self.zi * self.DPhi - self.inv_X]

        self.grad_updated = True

    def update_hessprod_aux(self):
        assert not self.hess_aux_updated
        assert self.grad_updated

        self.zi2 = self.zi * self.zi

        self.D1gx_log_blk = [
            D1_log(D, log_D) for (D, log_D) in zip(self.Dgx_blk, self.log_Dgx_blk)
        ]
        self.D1zgx_log_blk = [
            D1_log(D, log_D) for (D, log_D) in zip(self.Dzgx_blk, self.log_Dzgx_blk)
        ]

        if self.G_is_Id:
            D1x_inv = np.reciprocal(np.outer(self.Dx, self.Dx))
            self.D1x_comb = self.zi * self.D1gx_log_blk[0] + D1x_inv

        self.hess_aux_updated = True

        return

    def hess_prod_ip(self, out, H):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()

        (Ht, Hx) = H

        K_list_blk, ZK_list_blk = self.K_list_blk, self.ZK_list_blk
        Ugx_blk, Uzgx_blk = self.Ugx_blk, self.Uzgx_blk
        D1gx_log_blk, D1zgx_log_blk = self.D1gx_log_blk, self.D1zgx_log_blk

        GH_blk = [apply_kraus(Hx, K_list) for K_list in K_list_blk]
        ZGH_blk = [apply_kraus(Hx, ZK_list) for ZK_list in ZK_list_blk]

        UGHU_blk = [U.conj().T @ H @ U for (H, U) in zip(GH_blk, Ugx_blk)]
        UZGHU_blk = [U.conj().T @ H @ U for (H, U) in zip(ZGH_blk, Uzgx_blk)]

        # Hessian products for quantum key distribution
        # D2Phi(X)[H]
        #     = G'(Ugx [log^[1](Dgx) .* (Ugx' G(H) Ugx)] Ugx')
        #       - G'(Z'(Uzgx [log^[1](Dzgx) .* (Uzgx' Z(G(H)) Uzgx)] Uzgx'))
        D2PhiH = np.zeros_like(self.X)
        # First term, i.e., G'(Ugx [log^[1](Dgx) .* (Ugx' G(H) Ugx)] Ugx')
        for k, (U_k, D1_k) in enumerate(zip(Ugx_blk, D1gx_log_blk)):
            temp = U_k @ (D1_k * UGHU_blk[k]) @ U_k.conj().T
            D2PhiH += apply_kraus(temp, K_list_blk[k], adjoint=True)
        # Second term, i.e., -G'(Z'(Uzgx [ ... ] Uzgx'))
        for k, (U_k, D1_k) in enumerate(zip(Uzgx_blk, D1zgx_log_blk)):
            temp = U_k @ (D1_k * UZGHU_blk[k]) @ U_k.conj().T
            D2PhiH -= apply_kraus(temp, ZK_list_blk[k], adjoint=True)

        # ======================================================================
        # Hessian products with respect to t
        # ======================================================================
        # D2_t F(t, X)[Ht, Hx] = (Ht - D_X Phi(X)[Hx]) / z^2
        out[0][:] = (Ht - inp(self.DPhi, Hx)) * self.zi2

        # ======================================================================
        # Hessian products with respect to X
        # ======================================================================
        # D2_X F(t, X)[Ht, Hx] = -D2_t F(t, X)[Ht, Hx] * DPhi(X)
        #                        + D2Phi(X)[Hx] / z + X^-1 Hx X^-1
        out_X = -out[0] * self.DPhi
        out_X += self.zi * D2PhiH
        out_X += self.inv_X @ Hx @ self.inv_X
        out[1][:] = (out_X + out_X.conj().T) * 0.5

        return out

    def hess_congr(self, A):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.congr_aux_updated:
            self.congr_aux(A)
        if not self.hess_congr_aux_updated:
            self.update_hess_congr_aux(A)

        p = A.shape[0]
        lhs = np.zeros((p, sum(self.dim)))

        Z_idxs = self.Z_idxs
        K_list_blk, ZK_list_blk = self.K_list_blk, self.ZK_list_blk
        Ugx_blk, Uzgx_blk = self.Ugx_blk, self.Uzgx_blk
        D1gx_log_blk, D1zgx_log_blk = self.D1gx_log_blk, self.D1zgx_log_blk

        Work0, Work1 = self.Work0, self.Work1
        Work2, Work3 = self.Work2, self.Work3

        # ======================================================================
        # Hessian products with respect to t
        # ======================================================================
        # D2_t F(t, X)[Ht, Hx] = (Ht - D_X Phi(X)[Hx]) / z^2
        Ax_vec = self.Ax.view(np.float64).reshape((p, 1, -1))
        DPhi_vec = self.DPhi.view(np.float64).reshape((-1, 1))
        outt = self.At - (Ax_vec @ DPhi_vec).ravel()
        outt *= self.zi2

        lhs[:, 0] = outt

        # ======================================================================
        # Hessian products with respect to X
        # ======================================================================
        # Hessian products for quantum key distribution
        # D2Phi(X)[H]
        #     = G'(Ugx [log^[1](Dgx) .* (Ugx' G(H) Ugx)] Ugx')
        #       - G'(Z'(Uzgx [log^[1](Dzgx) .* (Uzgx' Z(G(H)) Uzgx)] Uzgx'))
        if self.G_is_Id:
            work1, work2, work3 = self.work1, self.work2, self.work2

            # If G(X) = X, then we can do some more efficient slicing operations
            # Compute second term, i.e., -Z'(Uzx [log^[1](Dzx) .* ... ] Uzx')
            Work0 *= 0
            for Z_idxs_k, U_k, D1_log_k in zip(Z_idxs, Uzgx_blk, D1zgx_log_blk):
                # Apply Z(Ax), i.e., extract k-th submatrix from Ax
                temp = self.Ax[np.ix_(np.arange(p), Z_idxs_k, Z_idxs_k)]
                # Compute Uzx [log^[1](Dzx) .* (Uzgx' Z(Ax) Uzx)] Uzx'
                congr_multi(work2, U_k.conj().T, temp, work3)
                work2 *= D1_log_k * self.zi
                congr_multi(work1, U_k, work2, work3)
                # Apply Z'( ... ), i.e., place result in k-th submatrix
                Work0[np.ix_(np.arange(p), Z_idxs_k, Z_idxs_k)] = work1

            # Compute first term, i.e., Ux [log^[1](Dx) .* (Ux' G(H) Ux)] Ux'
            # Also combine this with 1/z ( ... ) + X^-1 Ax X^-1
            congr_multi(Work2, self.Ux.conj().T, self.Ax, Work3)
            self.Work2 *= self.D1x_comb
            congr_multi(Work1, self.Ux, Work2, Work3)

            # Subtract two terms to get D2Phi(X)[H]
            Work1 -= Work0
        else:
            Work1 *= 0

            # Get first term, i.e., G'(Ugx [log^[1](Dgx) .* ... ] Ugx')
            for k in range(len(K_list_blk)):
                worka, workb = self.Work4[k], self.Work4b[k]
                workc, workd = self.Work5[k], self.Work5b[k]
                workc *= 0

                KU_list = [K.conj().T @ Ugx_blk[k] for K in K_list_blk[k]]
                # Apply Ugx' G(Ax) Ugx
                for KU in KU_list:
                    congr_multi(workd, KU.conj().T, self.Ax, work=workb)
                    workc += workd
                # Apply log^[1](Dgx) .* ( ... )
                workc *= D1gx_log_blk[k] * self.zi
                # Apply G'(Ugx [ ... ] Ugx')
                for KU in KU_list:
                    congr_multi(Work0, KU, workc, work=worka)
                    Work1 += Work0

            # Get second term, i.e., G'(Z'(Uzgx [log^[1](Dzgx) .* ... ] Uzgx'))
            for k in range(len(ZK_list_blk)):
                worka, workb = self.Work6[k], self.Work6b[k]
                workc, workd = self.Work7[k], self.Work7b[k]
                workc *= 0

                KU_list = [K.conj().T @ Uzgx_blk[k] for K in ZK_list_blk[k]]
                # Apply Uzgx' Z(G(H)) Uzgx
                for KU in KU_list:
                    congr_multi(workd, KU.conj().T, self.Ax, work=workb)
                    workc += workd
                # Apply log^[1](Dzgx) .* ( ... )
                workc *= D1zgx_log_blk[k] * self.zi
                # Apply G'(Z'(Uzgx [ ... ] Uzgx'))
                for KU in KU_list:
                    congr_multi(Work0, KU, workc, work=worka)
                    Work1 -= Work0

            # Compute X^-1 Ax X^-1
            congr_multi(Work0, self.inv_X, self.Ax, Work2)
            Work1 += Work0

        # Hessian product of barrier function
        # D2_X F(t, X)[Ht, Hx] = -D2_t F(t, X)[Ht, Hx] * DPhi(X)
        #                        + D2Phi(X)[Hx] / z + X^-1 Hx X^-1
        np.outer(outt, self.DPhi, out=Work0.reshape(p, -1))
        Work1 -= Work0

        lhs[:, 1:] = Work1.reshape((p, -1)).view(np.float64)

        # Multiply A (H A')
        return dense_dot_x(lhs, A.T)

    def invhess_prod_ip(self, out, H):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.invhess_aux_updated:
            self.update_invhessprod_aux()

        (Ht, Hx) = H

        Wx = Hx + Ht * self.DPhi

        # ======================================================================
        # Inverse Hessian products with respect to X
        # ======================================================================
        # Compute X = M \ Wx, where we solve this using two different strategies
        if self.G_is_Id:
            # If G(X) = X, then solve M using the matrix inversion lemma
            work = np.zeros((self.r * self.vm))

            # Apply D2S(X)^-1 = (Ux⊗Ux) (1/z log + inv)^[1](Dx)^-1 (Ux'⊗Ux')
            temp = self.Ux.conj().T @ Wx @ self.Ux
            out_X = self.Ux @ (self.D1x_comb_inv * temp) @ self.Ux.conj().T
            # Apply Z( ... ), i.e., extract k-th submatrix then get compact
            # vectorization of this submatrix
            for k, Z_idxs_k in enumerate(self.Z_idxs):
                temp_k = out_X[np.ix_(Z_idxs_k, Z_idxs_k)]
                temp_vec = temp_k.view(np.float64).reshape(-1)
                work[k * self.vm : (k + 1) * self.vm] = self.F2C_op @ temp_vec
            work *= -1
            # Solve linear system N \ ( ... )
            work = cho_solve(self.schur_fact, work)
            # Apply Z'( ... ), i.e., recover submatrix from compact
            # vectorization then place in k-th submatrix
            Work3 = np.zeros((self.n, self.n), dtype=self.dtype)
            for k, Z_idxs_k in enumerate(self.Z_idxs):
                work_k = work[k * self.vm : (k + 1) * self.vm]
                work_k = self.F2C_op.T @ work_k
                work_k = work_k.view(self.dtype).reshape((self.m, self.m))
                Work3[np.ix_(Z_idxs_k, Z_idxs_k)] = work_k

            # Apply D2S(X)^-1 = (Ux ⊗ Ux) log^[1](Dx) (Ux' ⊗ Ux') and subtract
            # from (D2S(X)^-1 Wx) to get X
            temp = self.Ux.conj().T @ Work3 @ self.Ux
            out_X -= self.Ux @ (self.D1x_comb_inv * temp) @ self.Ux.conj().T

        else:
            # Otherwise, directly build M and Cholesky factor to solve M \ Wx
            # Convert matrices to compact real vectors
            temp_vec = Wx.view(np.float64).reshape((-1, 1))
            temp_vec = self.F2C_op @ temp_vec
            # Solve system
            temp_vec = cho_solve(self.hess_fact, temp_vec)
            # Expand compact real vectors back into full matrices
            temp_vec = self.F2C_op.T @ temp_vec
            out_X = temp_vec.T.view(self.dtype).reshape((self.n, self.n))

        out[1][:] = (out_X + out_X.conj().T) * 0.5

        # ======================================================================
        # Inverse Hessian products with respect to t
        # ======================================================================
        # Compute t = z^2 Ht + <DPhi(X), X>
        out[0][:] = self.z2 * Ht + inp(self.DPhi, out[1])

        return out

    def invhess_congr(self, A):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.invhess_aux_updated:
            self.update_invhessprod_aux()
        if not self.congr_aux_updated:
            self.congr_aux(A)
        if not self.invhess_congr_aux_updated:
            self.update_invhess_congr_aux(A)

        # The inverse Hessian product applied on (Ht, Hx, Hy) for the OPT
        # barrier is
        #     X = M \ Wx
        #     t = z^2 Ht + <DPhi(X), X>
        # where Wx = Hx + Ht DPhi(X) and M = 1/z D2Phi + X^1 ⊗ X^-1

        if self.G_is_Id:
            # When G(X) = X,  we can write the expression for M as
            #   M = 1/z D2S(X) - 1/z Z' D2S(Z(X)) Z + X^-1 ⊗ X^-1
            #     = (Ux ⊗ Ux) (1/z log + inv)^[1](Dx) (Ux' ⊗ Ux')
            #       - 1/z Z' (Uzx ⊗ Uzx) log^[1](Dzx) (Uzx' ⊗ Uzx') Z
            # Treating [Z' D2S(Z(X)) Z] as a low-rank perturbation of D2S(X), we
            # can solve linear systems with M by using the matrix inversion
            # lemma
            #   X = [D2S(X)^-1 - D2S(X)^-1 Z' N^-1 Z D2S(X)^-1] Wx
            # where
            #   N = 1/z (Uzx ⊗ Uzx) [log^[1](Dzx)]^-1 (Uzx' ⊗ Uzx')
            #       - Z (Ux ⊗ Ux) [(1/z log + inv)^[1](Dx)]^-1  (Ux' ⊗ Ux') Z'

            p = A.shape[0]
            lhs = np.zeros((p, sum(self.dim)))

            work = np.zeros((self.r * self.vm, p))

            Work0, Work1 = self.Work0, self.Work1
            Work2, Work3 = self.Work2, self.Work3

            # ==================================================================
            # Inverse Hessian products with respect to X
            # ==================================================================
            # Compute Wx = Hx + Ht DPhi(X)
            np.outer(self.At, self.DPhi, out=Work2.reshape((p, -1)))
            np.add(self.Ax, Work2, out=Work0)

            # Apply D2S(X)^-1 = (Ux⊗Ux) (1/z log + inv)^[1](Dx)^-1 (Ux'⊗Ux')
            congr_multi(Work2, self.Ux.conj().T, Work0, Work3)
            Work2 *= self.D1x_comb_inv
            congr_multi(Work0, self.Ux, Work2, Work3)

            # Apply Z( ... ), i.e., extract k-th submatrix then get compact
            # vectorization of this submatrix
            for k, Z_idxs_k in enumerate(self.Z_idxs):
                temp = Work0[np.ix_(np.arange(p), Z_idxs_k, Z_idxs_k)]
                temp = temp.view(np.float64).reshape((p, -1)).T
                temp = x_dot_dense(self.F2C_op, temp)
                work[k * self.vm : (k + 1) * self.vm] = temp
            work *= -1
            # Solve linear system N \ ( ... )
            work = cho_solve(self.schur_fact, work)
            # Apply Z'( ... ), i.e., recover submatrix from compact
            # vectorization then place in k-th submatrix
            Work1.fill(0.0)
            for k, Z_idxs_k in enumerate(self.Z_idxs):
                work_k = work[k * self.vm : (k + 1) * self.vm]
                work_k = x_dot_dense(self.F2C_op.T, work_k)
                work_k = work_k.T.view(self.dtype).reshape(p, self.m, self.m)
                Work1[np.ix_(np.arange(p), Z_idxs_k, Z_idxs_k)] = work_k

            # Apply D2S(X)^-1 = (Ux⊗Ux) (1/z log + inv)^[1](Dx)^-1 (Ux'⊗Ux')
            congr_multi(Work2, self.Ux.conj().T, Work1, Work3)
            Work2 *= self.D1x_comb_inv
            congr_multi(Work1, self.Ux, Work2, Work3)

            # Subtract previous expression from (D2S(X)^-1 Wx) to get X
            Work0 -= Work1
            out_X = Work0.reshape((p, -1)).view(np.float64)
            lhs[:, 1:] = out_X

            # ==================================================================
            # Inverse Hessian products with respect to t
            # ==================================================================
            # Compute t = z^2 Ht + <DPhi(X), X>
            DPhi_vec = self.DPhi.view(np.float64).reshape((-1, 1))

            out_t = self.z2 * self.At
            out_t += (out_X @ DPhi_vec).ravel()
            lhs[:, 0] = out_t

            # Multiply A (H A')
            return dense_dot_x(lhs, A.T)

        else:
            # Otherwise, we will directly build M and Cholesky factor it to
            # solve X = M \ Wx

            # ==================================================================
            # Inverse Hessian products with respect to X
            # ==================================================================
            # Compute Wx = Hx + Ht DPhi(X)
            np.outer(self.At, self.DPhi_cvec, out=self.work0)
            self.work0 += self.Ax_cvec

            # Solve system X = M \ Wx
            out_X = cho_solve(self.hess_fact, self.work0.T)

            # ==================================================================
            # Inverse Hessian products with respect to t
            # ==================================================================
            # Compute t = z^2 Ht + <DPhi(X), X>
            out_t = self.z2 * self.At
            out_t += (out_X.T @ self.DPhi_cvec).ravel()

            # Multiply A (H A')
            out = np.outer(out_t, self.At)
            out += x_dot_dense(self.Ax_cvec, out_X)
            return out

    def third_dir_deriv_axpy(self, out, H, a=True):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.dder3_aux_updated:
            self.update_dder3_aux()

        (Ht, Hx) = H

        K_list_blk, ZK_list_blk = self.K_list_blk, self.ZK_list_blk
        Ugx_blk, Uzgx_blk = self.Ugx_blk, self.Uzgx_blk
        D1gx_log_blk, D1zgx_log_blk = self.D1gx_log_blk, self.D1zgx_log_blk
        D2gx_log_blk, D2zgx_log_blk = self.D2gx_log_blk, self.D2zgx_log_blk

        GH_blk = [apply_kraus(Hx, K_list) for K_list in K_list_blk]
        ZGH_blk = [apply_kraus(Hx, ZK_list) for ZK_list in ZK_list_blk]

        UGHU_blk = [U.conj().T @ H @ U for (H, U) in zip(GH_blk, Ugx_blk)]
        UZGHU_blk = [U.conj().T @ H @ U for (H, U) in zip(ZGH_blk, Uzgx_blk)]

        # Quantum key distribution Hessians
        D2PhiH = np.zeros_like(self.X)
        # Hessians of S(G(X))
        for k, (U_k, D1_k) in enumerate(zip(Ugx_blk, D1gx_log_blk)):
            temp = U_k @ (D1_k * UGHU_blk[k]) @ U_k.conj().T
            D2PhiH += apply_kraus(temp, K_list_blk[k], adjoint=True)
        # Hessians of S(Z(G(X)))
        for k, (U_k, D1_k) in enumerate(zip(Uzgx_blk, D1zgx_log_blk)):
            temp = U_k @ (D1_k * UZGHU_blk[k]) @ U_k.conj().T
            D2PhiH -= apply_kraus(temp, ZK_list_blk[k], adjoint=True)

        # Quantum key distribution third order derivatives
        D3PhiHH = np.zeros_like(self.X)
        # Third order derivatives of S(G(X))
        for k, (U_k, D2_k) in enumerate(zip(Ugx_blk, D2gx_log_blk)):
            temp = scnd_frechet(D2_k * UGHU_blk[k], UGHU_blk[k], U=U_k)
            D3PhiHH += apply_kraus(temp, K_list_blk[k], adjoint=True)
        # Third order derivatives of S(Z(G(X)))
        for k, (U_k, D2_k) in enumerate(zip(Uzgx_blk, D2zgx_log_blk)):
            temp = scnd_frechet(D2_k * UZGHU_blk[k], UZGHU_blk[k], U=U_k)
            D3PhiHH -= apply_kraus(temp, ZK_list_blk[k], adjoint=True)

        # Third derivative of barrier
        DPhiH = inp(self.DPhi, Hx)
        D2PhiHH = inp(D2PhiH, Hx)
        chi = Ht - DPhiH
        chi2 = chi * chi

        dder3_t = -2 * self.zi3 * chi2 - self.zi2 * D2PhiHH

        dder3_X = -dder3_t * self.DPhi
        dder3_X -= 2 * self.zi2 * chi * D2PhiH
        dder3_X += self.zi * D3PhiHH
        dder3_X -= 2 * self.inv_X @ Hx @ self.inv_X @ Hx @ self.inv_X
        dder3_X = (dder3_X + dder3_X.conj().T) * 0.5

        out[0][:] += dder3_t * a
        out[1][:] += dder3_X * a

        return out

    # ========================================================================
    # Auxilliary functions
    # ========================================================================
    def congr_aux(self, A):
        assert not self.congr_aux_updated

        iscomplex = self.iscomplex

        # Get slices and views of A matrix to be used in congruence computations
        if sp.sparse.issparse(A):
            A = A.tocsr()
        self.Ax_vec = A[:, 1:]

        if sp.sparse.issparse(A):
            A = A.toarray()
        Ax_dense = np.ascontiguousarray(A[:, 1:])
        self.At = A[:, 0]
        self.Ax = np.array([vec_to_mat(Ax_k, iscomplex) for Ax_k in Ax_dense])

        self.congr_aux_updated = True

    def update_hess_congr_aux(self, A):
        assert not self.hess_congr_aux_updated

        p = A.shape[0]

        self.Work0 = np.zeros_like(self.Ax, dtype=self.dtype)
        self.Work1 = np.zeros_like(self.Ax, dtype=self.dtype)
        self.Work2 = np.zeros_like(self.Ax, dtype=self.dtype)
        self.Work3 = np.zeros_like(self.Ax, dtype=self.dtype)

        if self.G_is_Id:
            self.work1 = np.empty((p, self.m, self.m), dtype=self.dtype)
            self.work2 = np.empty((p, self.m, self.m), dtype=self.dtype)
            self.work3 = np.empty((p, self.m, self.m), dtype=self.dtype)
        else:
            n, Nk, Nzk, dtype = self.n, self.Nk, self.Nzk, self.dtype
            self.Work4 = [np.zeros((p, n, nk), dtype=dtype) for nk in Nk]
            self.Work4b = [np.zeros((p, nk, n), dtype=dtype) for nk in Nk]
            self.Work5 = [np.zeros((p, nk, nk), dtype=dtype) for nk in Nk]
            self.Work5b = [np.zeros((p, nk, nk), dtype=dtype) for nk in Nk]
            self.Work6 = [np.zeros((p, n, nzk), dtype=dtype) for nzk in Nzk]
            self.Work6b = [np.zeros((p, nzk, n), dtype=dtype) for nzk in Nzk]
            self.Work7 = [np.zeros((p, nzk, nzk), dtype=dtype) for nzk in Nzk]
            self.Work7b = [np.zeros((p, nzk, nzk), dtype=dtype) for nzk in Nzk]

        self.hess_congr_aux_updated = True

    def update_invhess_congr_aux(self, A):
        assert not self.invhess_congr_aux_updated

        if self.G_is_Id:
            self.Work0 = np.zeros_like(self.Ax, dtype=self.dtype)
            self.Work1 = np.zeros_like(self.Ax, dtype=self.dtype)
            self.Work2 = np.zeros_like(self.Ax, dtype=self.dtype)
            self.Work3 = np.zeros_like(self.Ax, dtype=self.dtype)
        else:
            if sp.sparse.issparse(A):
                A = A.tocsr()
            self.Ax_cvec = (self.F2C_op @ A[:, 1:].T).T
            if sp.sparse.issparse(A):
                self.Ax_cvec = self.Ax_cvec.tocoo()

            self.work0 = np.zeros(self.Ax_cvec.shape)
            self.work1 = np.zeros(self.Ax_cvec.shape)

        self.invhess_congr_aux_updated = True

    def update_invhessprod_aux_aux(self):
        assert not self.invhess_aux_aux_updated

        if self.G_is_Id:
            self.work6 = np.zeros((self.vm, self.m, self.m), dtype=self.dtype)
            self.work7 = np.zeros((self.vm, self.m, self.m), dtype=self.dtype)
            self.work8 = np.zeros((self.vm, self.m, self.m), dtype=self.dtype)

            # Computational basis for symmetric/Hermitian matrices
            rt2 = np.sqrt(0.5)
            n, r, m, vm = self.n, self.r, self.m, self.vm
            self.E_blk = np.zeros((r, vm, n, n), dtype=self.dtype)
            for b, Z_idxs_k in enumerate(self.Z_idxs):
                k = 0
                for j_subblk in range(m):
                    j = Z_idxs_k[j_subblk]
                    for i_subblk in range(j_subblk):
                        i = Z_idxs_k[i_subblk]
                        self.E_blk[b, k, i, j] = rt2
                        self.E_blk[b, k, j, i] = rt2
                        k += 1
                        if self.iscomplex:
                            self.E_blk[b, k, i, j] = rt2 * 1j
                            self.E_blk[b, k, j, i] = rt2 * -1j
                            k += 1
                    self.E_blk[b, k, j, j] = 1.0
                    k += 1
            self.E_blk = self.E_blk.reshape((-1, self.n, self.n))

            self.Work6 = np.zeros((r * vm, n, n), dtype=self.dtype)
            self.Work7 = np.zeros((r * vm, n, n), dtype=self.dtype)
            self.Work8 = np.zeros((r * vm, n, n), dtype=self.dtype)
        else:
            dtype = self.dtype
            n, vn, Nk, Nzk = self.n, self.vn, self.Nk, self.Nzk
            self.work2 = [np.zeros((vn, n, nk), dtype=dtype) for nk in Nk]
            self.work2b = [np.zeros((vn, nk, n), dtype=dtype) for nk in Nk]
            self.work3 = [np.zeros((vn, nk, nk), dtype=dtype) for nk in Nk]
            self.work3b = [np.zeros((vn, nk, nk), dtype=dtype) for nk in Nk]
            self.work4 = [np.zeros((vn, n, nzk), dtype=dtype) for nzk in Nzk]
            self.work4b = [np.zeros((vn, nzk, n), dtype=dtype) for nzk in Nzk]
            self.work5 = [np.zeros((vn, nzk, nzk), dtype=dtype) for nzk in Nzk]
            self.work5b = [np.zeros((vn, nzk, nzk), dtype=dtype) for nzk in Nzk]
            self.work6 = np.zeros((vn, n, n), dtype=dtype)
            self.work7 = np.zeros((vn, n, n), dtype=dtype)
            self.work8 = np.zeros((vn, n, n), dtype=dtype)

        self.invhess_aux_aux_updated = True

    def update_invhessprod_aux(self):
        assert not self.invhess_aux_updated
        assert self.grad_updated
        assert self.hess_aux_updated
        if not self.invhess_aux_aux_updated:
            self.update_invhessprod_aux_aux()

        self.z2 = self.z * self.z

        if self.G_is_Id:
            # Precompute and factorize the matrix
            #   N = z (Uzx ⊗ Uzx) [log^[1](Dzx)]^-1 (Uzx' ⊗ Uzx')
            #       - Z (Ux ⊗ Ux) [(1/z log + inv)^[1](Dx)]^-1  (Ux' ⊗ Ux') Z'
            # which we will need to solve linear systems with the Hessian of
            # our barrier function

            self.D1x_comb_inv = np.reciprocal(self.D1x_comb)

            r, vm = self.r, self.vm
            Uzgx_blk, D1zgx_log_blk = self.Uzgx_blk, self.D1zgx_log_blk

            work6, work7, work8 = self.work6, self.work7, self.work8
            Work6, Work7, Work8 = self.Work6, self.Work7, self.Work8

            self.schur = np.zeros((r * vm, r * vm))

            # ==================================================================
            # Get first term, i.e., 1/z (Uzx ⊗ Uzx) ... (Uzx' ⊗ Uzx')
            # ==================================================================
            for k, (U, D1_log) in enumerate(zip(Uzgx_blk, D1zgx_log_blk)):
                # Begin with (Uzx' ⊗ Uzx')
                congr_multi(work8, U.conj().T, self.E, work=work7)
                # Apply z [log^[1](Dzx)]^-1
                work8 *= self.z * np.reciprocal(D1_log)
                # Apply (Uzx ⊗ Uzx)
                congr_multi(work6, U, work8, work=work7)

                work = work6.view(np.float64).reshape((vm, -1))
                work = x_dot_dense(self.F2C_op, work.T)

                self.schur[k * vm : (k + 1) * vm, k * vm : (k + 1) * vm] = work

            # ==================================================================
            # Get second term, i.e., Z (Ux ⊗ Ux) ...  (Ux' ⊗ Ux') Z'
            # ==================================================================
            # Begin with (Ux' ⊗ Ux') Z'
            congr_multi(Work8, self.Ux.conj().T, self.E_blk, work=Work7)
            # Apply [(1/z log + inv)^[1](Dx)]^-1
            Work8 *= self.D1x_comb_inv
            # Apply (Ux ⊗ Ux)
            congr_multi(Work6, self.Ux, Work8, work=Work7)
            # Apply Z, i.e., extract submatrices
            for k, Z_idxs_k in enumerate(self.Z_idxs):
                temp = Work6[np.ix_(np.arange(r * vm), Z_idxs_k, Z_idxs_k)]
                temp = temp.view(np.float64).reshape((r * vm, -1))
                temp = x_dot_dense(self.F2C_op, temp.T).T

                self.schur[:, k * vm : (k + 1) * vm] -= temp

            # Subtract terms to obtain N then Cholesky factor
            self.schur_fact = cho_fact(self.schur)

        else:
            # Precompute and factorize the matrix
            #   M = 1/z G' (Ugx ⊗ Ugx) log^[1](Dgx) (Ugx' ⊗ Ugx') G
            #       - 1/z G'Z' (Uzgx ⊗ Uzgx) log^[1](Dzgx) (Uzgx' ⊗ Uzgx') ZG
            #       +  X^-1 ⊗ X^-1
            # which we will need to solve linear systems with the Hessian of
            # our barrier function

            K_list_blk, ZK_list_blk = self.K_list_blk, self.ZK_list_blk
            Ugx_blk, Uzgx_blk = self.Ugx_blk, self.Uzgx_blk
            D1gx_log_blk, D1zgx_log_blk = self.D1gx_log_blk, self.D1zgx_log_blk

            work7, work8 = self.work7, self.work8

            DPhi_vec = self.DPhi.view(np.float64).reshape(-1, 1)
            self.DPhi_cvec = self.F2C_op @ DPhi_vec

            # ==================================================================
            # Get third term, i.e., X^-1 ⊗ X^-1
            # ==================================================================
            congr_multi(work8, self.inv_X, self.E, work=work7)

            # ==================================================================
            # Get first term, i.e., 1/z G' (Ugx ⊗ Ugx) ... (Ugx' ⊗ Ugx') G
            # ==================================================================
            for k in range(len(K_list_blk)):
                worka, workb = self.work2[k], self.work2b[k]
                workc, workd = self.work3[k], self.work3b[k]
                workc *= 0

                KU_list = [K.conj().T @ Ugx_blk[k] for K in K_list_blk[k]]
                # Apply Ugx' G(Ax) Ugx
                for KU in KU_list:
                    congr_multi(workd, KU.conj().T, self.E, work=workb)
                    workc += workd
                # Apply log^[1](Dgx) .* ( ... )
                workc *= D1gx_log_blk[k] * self.zi
                # Apply G'(Ugx [ ... ] Ugx')
                for KU in KU_list:
                    congr_multi(work7, KU, workc, work=worka)
                    work8 += work7

            # ==================================================================
            # Get second term, i.e., 1/z G'Z' [ ... ] ZG
            # ==================================================================
            for k in range(len(ZK_list_blk)):
                worka, workb = self.work4[k], self.work4b[k]
                workc, workd = self.work5[k], self.work5b[k]
                workc *= 0

                KU_list = [K.conj().T @ Uzgx_blk[k] for K in ZK_list_blk[k]]
                # Apply Uzgx' Z(G(Ax)) Uzgx
                for KU in KU_list:
                    congr_multi(workd, KU.conj().T, self.E, work=workb)
                    workc += workd
                # Apply log^[1](Dzgx) .* ( ... )
                workc *= D1zgx_log_blk[k] * self.zi
                # Apply G'Z'((Uzgx [ ... ] Uzgx'))
                for KU in KU_list:
                    congr_multi(work7, KU, workc, work=worka)
                    work8 -= work7

            # Get Hessian and factorize
            work = work8.view(np.float64).reshape((self.vn, -1))
            self.hess = x_dot_dense(self.F2C_op, work.T)
            self.hess = (self.hess + self.hess.T) * 0.5
            self.hess_fact = cho_fact(self.hess)

        self.invhess_aux_updated = True

    def update_dder3_aux(self):
        assert not self.dder3_aux_updated
        assert self.hess_aux_updated

        self.zi3 = self.zi2 * self.zi

        self.D2gx_log_blk = [
            D2_log(D, D1) for (D, D1) in zip(self.Dgx_blk, self.D1gx_log_blk)
        ]
        self.D2zgx_log_blk = [
            D2_log(D, D1) for (D, D1) in zip(self.Dzgx_blk, self.D1zgx_log_blk)
        ]

        self.dder3_aux_updated = True

        return


def facial_reduction(K_list):
    # For a set of Kraus operators i.e., Σ_i K_i @ X @ K_i.T, returns a set of
    # Kraus operators which preserves positive definiteness
    nk = K_list[0].shape[0]

    # Pass identity matrix (maximally mixed state) through the Kraus operators
    KK = sum([K @ K.conj().T for K in K_list])

    # Determine if output is low rank, in which case we need to perform facial
    # reduction
    Dkk, Ukk = np.linalg.eigh(KK)
    KKnzidx = np.where(Dkk > 1e-12)[0]
    nk_fr = np.size(KKnzidx)

    if nk == nk_fr:
        return K_list

    # Perform facial reduction
    Qkk = Ukk[:, KKnzidx]
    K_list_fr = [Qkk.conj().T @ K for K in K_list]

    return K_list_fr


def apply_kraus(x, Klist, adjoint=False):
    # Compute congruence map
    if adjoint:
        return sum([K.conj().T @ x @ K for K in Klist])
    else:
        return sum([K @ x @ K.conj().T for K in Klist])



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

from qics.point import Point, VecProduct


class SymStepper:
    def __init__(self, kktsolver, model, toa=True):
        self.toa = toa
        if not toa:
            self.dir_dummy = VecProduct(model.cones)

        self.kktsolver = kktsolver

        self.rhs = Point(model)
        self.dir_a = Point(model)
        self.dir_comb = Point(model)
        self.next_point = Point(model)

        return

    def step(self, model, point, res, mu, verbose):
        # Step 1: Pre-build and -factor Schur complement matrix
        self.kktsolver.update_lhs(model, point, mu)

        # Step 2: Get affine direction
        self.update_rhs_affine(model, point, res)
        res_norm = self.kktsolver.solve_sys(self.dir_a, self.rhs, ir=False)

        # Step 3: Step size and centering parameter
        alpha = self.taukap_step_to_boundary(self.dir_a, point)
        for k, cone_k in enumerate(model.cones):
            alpha = min(
                alpha, cone_k.step_to_boundary(self.dir_a.s[k], self.dir_a.z[k])
            )
        alpha = min(alpha, 1.0)
        sigma = (1 - alpha) ** 3

        # Step 4: Combined direction
        self.update_rhs_comb(model, point, mu, self.dir_a, sigma, res)
        temp_res_norm = self.kktsolver.solve_sys(self.dir_comb, self.rhs, ir=True)
        res_norm = max(temp_res_norm, res_norm)

        # Step 5: Line search
        alpha = self.taukap_step_to_boundary(self.dir_comb, point)
        for k, cone_k in enumerate(model.cones):
            alpha = min(
                alpha, cone_k.step_to_boundary(self.dir_comb.s[k], self.dir_comb.z[k])
            )
        alpha = min(alpha, 1.0) * 0.99
        if alpha == 0:
            point, False

        if verbose == 3:
            print(f"  |  {res_norm:>7.1e}   {sigma:>5.3f}   {alpha:>5.3f}", end="")

        # Take step
        point.vec += alpha * self.dir_comb.vec
        for k, cone_k in enumerate(model.cones):
            cone_k.set_point(point.s[k], point.z[k])
            if not cone_k.get_feas():
                return point, False, alpha

        return point, True, alpha

    def taukap_step_to_boundary(self, dir, point):
        tau_alpha = dir.tau[0, 0] / point.tau[0, 0]
        kap_alpha = dir.kap[0, 0] / point.kap[0, 0]
        if tau_alpha >= 0 and kap_alpha >= 0:
            # Can take maximum step
            return 1.0
        else:
            return 1.0 / max(-tau_alpha, -kap_alpha)

    def update_rhs_affine(self, model, point, res):
        self.rhs.x[:] = res["x"]
        self.rhs.y[:] = res["y"]
        self.rhs.z.vec[:] = res["z"]

        self.rhs.s.vec[:] = -1 * point.z.vec

        self.rhs.tau[:] = res["tau"]
        self.rhs.kap[:] = -point.kap

        return self.rhs

    def update_rhs_comb(self, model, point, mu, dir_a, sigma, res):
        self.rhs.x[:] = res["x"] * (1 - sigma)
        self.rhs.y[:] = res["y"] * (1 - sigma)
        self.rhs.z.vec[:] = res["z"] * (1 - sigma)

        if self.toa:
            for k, cone_k in enumerate(model.cones):
                cone_k.comb_dir(self.rhs.s[k], dir_a.s[k], dir_a.z[k], sigma * mu)
        else:
            dir_dummy = self.dir_dummy  # Represents the zero vector to avoid TOA
            for k, cone_k in enumerate(model.cones):
                cone_k.comb_dir(self.rhs.s[k], dir_dummy[k], dir_dummy[k], sigma * mu)

        self.rhs.tau[:] = res["tau"] * (1 - sigma)
        self.rhs.kap[:] = -point.kap + (-dir_a.kap * dir_a.tau + sigma * mu) / point.tau

        return self.rhs



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import numpy as np

from qics._stepper.kktsolver import blk_hess_congruence
from qics._utils.linalg import cho_fact, cho_solve
from qics.point import Point, VecProduct

alpha_sched = [0.9999, 0.999, 0.99, 0.9, 0.8, 0.7, 0.5, 0.3, 0.2, 0.1, 0.01, 0.001]


class NonSymStepper:
    def __init__(self, kktsolver, model, toa=True):
        self.toa = toa

        self.kktsolver = kktsolver
        self.prox = 0.0

        self.rhs = Point(model)
        self.dir_c = Point(model)
        self.dir_c_toa = Point(model)
        self.dir_p = Point(model)
        self.dir_p_toa = Point(model)
        self.next_point = Point(model)

        self.grad = VecProduct(model.cones)

        return

    def step(self, model, point, res, mu, verbose):
        self.kktsolver.update_lhs(model, point, mu)

        # Get prediction direction
        self.update_rhs_pred(model, point, res)
        res_norm = self.kktsolver.solve_sys(self.dir_p, self.rhs)

        # Get TOA prediction direction
        if self.toa:
            self.update_rhs_pred_toa(model, point, mu, self.dir_p)
            temp_res_norm = self.kktsolver.solve_sys(self.dir_p_toa, self.rhs)
            res_norm = max(temp_res_norm, res_norm)

        # Get centering direction
        self.update_rhs_cent(model, point, mu)
        temp_res_norm = self.kktsolver.solve_sys(self.dir_c, self.rhs)
        res_norm = max(temp_res_norm, res_norm)

        # Get TOA centering direction
        if self.toa:
            self.update_rhs_cent_toa(model, point, mu, self.dir_c)
            temp_res_norm = self.kktsolver.solve_sys(self.dir_c_toa, self.rhs)
            res_norm = max(temp_res_norm, res_norm)

        success = False

        if self.toa:
            step_mode = "co_toa"
            point, alpha, success = self.line_search(model, point, step_mode)

        if not success:
            step_mode = "comb"
            point, alpha, success = self.line_search(model, point, step_mode)

        if not success and self.toa:
            step_mode = "ce_toa"
            point, alpha, success = self.line_search(model, point, step_mode)

        if not success:
            step_mode = "cent"
            point, alpha, success = self.line_search(model, point, step_mode)

        if verbose == 3:
            print(
                f"  |  {step_mode:>6}   {res_norm:>7.1e}   {self.prox:>7.1e}   {alpha:>5.3f}",
                end="",
            )

        return point, success, alpha

    def line_search(self, model, point, mode="co_toa"):
        alpha = alpha_sched[0]
        alpha_iter = -1
        eta = 0.99

        dir_p = self.dir_p
        dir_c = self.dir_c
        dir_p_toa = self.dir_p_toa
        dir_c_toa = self.dir_c_toa
        next_point = self.next_point

        while True:
            alpha_iter += 1

            if alpha_iter >= len(alpha_sched):
                return point, alpha, False

            alpha = alpha_sched[alpha_iter]

            # Step point in direction and step size
            next_point.vec[:] = point.vec
            if mode == "co_toa":
                # step := alpha * (dir_p + dir_p_toa * alpha) + (dir_c + dir_c_toa * (1.0 - alpha)) * (1.0 - alpha)
                next_point.axpy(alpha, dir_p)
                next_point.axpy(alpha**2, dir_p_toa)
                next_point.axpy((1.0 - alpha), dir_c)
                next_point.axpy((1.0 - alpha) ** 2, dir_c_toa)
            elif mode == "comb":
                # step := dir_p * alpha + dir_c * (1.0 - alpha)
                next_point.axpy(alpha, dir_p)
                next_point.axpy((1.0 - alpha), dir_c)
            elif mode == "ce_toa":
                # step := (dir_p + dir_p_toa * alpha) * alpha
                next_point.axpy(alpha, dir_c)
                next_point.axpy(alpha**2, dir_c_toa)
            elif mode == "cent":
                # step := dir_c * alpha
                next_point.axpy(alpha, dir_c)

            # Check that tau, kap, mu are well defined
            # Make sure tau, kap are positive
            taukap = next_point.tau * next_point.kap
            if next_point.tau <= 0 or next_point.kap <= 0 or taukap <= 0:
                continue
            # Compute barrier parameter mu, ensure it is positive
            sz = next_point.s.inp(next_point.z)
            mu = (sz + taukap) / model.nu
            if sz <= 0 or mu <= 0:
                continue

            # Check cheap proximity conditions first
            szs = [
                s_k.T @ z_k for (s_k, z_k) in zip(next_point.s.vecs, next_point.z.vecs)
            ]
            nus = [cone_k.nu for cone_k in model.cones]
            rho = [
                np.abs(sz_k / mu - nu_k) / np.sqrt(nu_k)
                for (sz_k, nu_k) in zip(szs, nus)
            ]
            if any(np.array(rho) > eta):
                continue

            rtmu = np.sqrt(mu)
            irtmu = np.reciprocal(rtmu)
            self.prox = 0.0

            # Check feasibility
            in_prox = True
            for k, cone_k in enumerate(model.cones):
                cone_k.set_point(next_point.s[k], next_point.z[k], irtmu)

                # Check if feasible
                if not cone_k.get_feas():
                    in_prox = False
                    break

                if self.kktsolver.use_invhess:
                    # Check if close enough to central path
                    prox_k = cone_k.prox()
                    self.prox = max(self.prox, prox_k)
                    in_prox = self.prox < eta
                    if not in_prox:
                        break
                else:
                    cone_k.grad_ip(self.grad[k])

            if in_prox and not self.kktsolver.use_invhess:
                GHG = blk_hess_congruence(model.G_T_views, model)
                self.kktsolver.GHG_fact = cho_fact(GHG)
                psi = model.G_T @ (next_point.z.vec / rtmu + self.grad.vec)

                self.prox = (psi.T @ cho_solve(self.kktsolver.GHG_fact, psi))[0, 0]
                in_prox = self.prox < eta

            # If feasible, return point
            if in_prox:
                point.vec[:] = next_point.vec
                return point, alpha, True

    def update_rhs_cent(self, model, point, mu):
        self.rhs.x.fill(0.0)
        self.rhs.y.fill(0.0)
        self.rhs.z.fill(0.0)

        # rs := -z - mu*g(s)
        rtmu = np.sqrt(mu)
        for k, cone_k in enumerate(model.cones):
            cone_k.grad_ip(self.rhs.s[k])
        self.rhs.s.vec *= -rtmu
        self.rhs.s.vec -= point.z.vec

        self.rhs.tau[:] = 0.0
        self.rhs.kap[:] = -point.kap + mu / point.tau

        return self.rhs

    def update_rhs_cent_toa(self, model, point, mu, dir_c):
        self.rhs.x.fill(0.0)
        self.rhs.y.fill(0.0)
        self.rhs.z.fill(0.0)

        rtmu = np.sqrt(mu)
        self.rhs.s.fill(0.0)
        for k, cone_k in enumerate(model.cones):
            cone_k.third_dir_deriv_axpy(self.rhs.s[k], dir_c.s[k])
        self.rhs.s.vec *= -0.5 / rtmu

        self.rhs.tau[:] = 0.0
        self.rhs.kap[:] = 0.0

        return self.rhs

    def update_rhs_pred(self, model, point, res):
        self.rhs.x[:] = res["x"]
        self.rhs.y[:] = res["y"]
        self.rhs.z.vec[:] = res["z"]

        self.rhs.s.vec[:] = -point.z.vec

        self.rhs.tau[:] = res["tau"]
        self.rhs.kap[:] = -point.kap

        return self.rhs

    def update_rhs_pred_toa(self, model, point, mu, dir_p):
        self.rhs.x.fill(0.0)
        self.rhs.y.fill(0.0)
        self.rhs.z.fill(0.0)

        rtmu = np.sqrt(mu)
        for k, cone_k in enumerate(model.cones):
            cone_k.hess_prod_ip(self.rhs.s[k], dir_p.s[k])
            cone_k.third_dir_deriv_axpy(self.rhs.s[k], dir_p.s[k], a=-0.5 / rtmu)

        self.rhs.tau[:] = 0.0
        self.rhs.kap[:] = -dir_p.tau * dir_p.kap / point.tau

        return self.rhs
